
;;; Support code for the transfer version of the Space Fortress task



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Change to production compilation thanks to Dan

(setf (gethash (vector 8 8) (comp-buffer-type-table (gethash 'perceptual *valid-compilation-buffer-types*))) t) 

(defun COMPOSE-PERCEPTUAL-BUFFER (p1 p1-s p2 p2-s buffer)
 (declare (ignore p2))
 (let* ((bn (intern (concatenate 'string (symbol-name buffer) ">")))
        (b= (intern (concatenate 'string "=" (symbol-name bn))))
        (b+ (intern (concatenate 'string "+" (symbol-name bn))))
        (b? (intern (concatenate 'string "?" (symbol-name bn))))

        (c1 (copy-tree (find b= (first p1-s) :key 'car)))
        (c2 (copy-tree (find b= (first p2-s) :key 'car)))
        (q1 (copy-tree (find b? (first p1-s) :key 'car)))
        (q2 (copy-tree (find b? (first p2-s) :key 'car)))

        (a1+ (copy-tree (find b+ (second p1-s) :key 'car)))
        (a2+ (copy-tree (find b+ (second p2-s) :key 'car))))

   (case (aif (cdr (assoc buffer (production-buffer-indices p1))) it 0)
     (0
      (list (append
             (when c2
               (list c2))
             (when q2
               (list q2)))
            (when a2+
              (list a2+))))
     ((4 12)
      (list (when c1
              (list c1))
            (when a1+
              (list a1+))))
     (8
      (list (append
             (awhen (buffer-condition-union c1 c2 nil)
                    (list it))
             (when q2
               (list q2)))
            (when a2+
              (list a2+))))
     ((16 24)
      (list (append
             (awhen (buffer-condition-union c1 c2 nil)
                    (list it))
             (when q1
               (list q1)))
            (when a2+
              (list a2+))))
     ((20 28)
      (list (append
             (when c1
               (list c1))
             (when q1
               (list q1)))
            (when a1+
              (list a1+)))))))

   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;global variables for monitoring state
  (defparameter *thrust* nil) ;has a thrust been issued
  (defparameter *vulnerability* 0); current vulnerability
  (defparameter *missiles* nil); number of missiles present
  (defparameter *points* 0)
  (defparameter *data* nil)
  (defparameter *rectangle1* nil)
  (defparameter *rectangle2* nil)
  (defparameter *turn* nil)

;;; This notes relevant information for purpose of updating the imaginal and game-state module in the two games
(defun record-features-fortress (data)  
;(push data *data*)
 (game-state-features-fortress data)
(if (equal (getf data :SCREEN-TYPE) "game")
  (let* ((vuln (getf data :vlner)) speed
         (collisions (getf data :COLLISIONS))
         (ship (getf data :ship))
         (keys (getf data :keys))
         (missiles (getf data :MISSILES))
         (points (getf data :pnts)))
    (if (and (numberp *points*) (> points *points*)) (update-slot  'kill))
    (cond ((member :thrust keys) (setf *thrust* t))
          ((and *thrust* ship) (setf *thrust* nil) (setf speed (speed-sf (getf ship :vx) (getf ship :vy)))
            (if (or (< speed 1) (> speed 1.7)) (update-slot 'badspeed) (update-slot 'goodspeed))))
    (cond ((intersection collisions '(:BIG-HEX)) (update-slot 'bighexdeath))
          ((intersection collisions '(:small-HEX))  (update-slot 'smallhexdeath))
          ((intersection collisions '(:shell))  (update-slot 'shelldeath)))
    (if (intersection collisions '(:shell :BIG-HEX :small-HEX))  (update-slot 'death))
    (if (and (numberp *vulnerability*)(numberp vuln) (= vuln 0) (member :FORTRESS collisions)  (< *vulnerability* 11)) (update-slot 'reset))
    (if (and (> (length *missiles*) (length missiles)) (not (and (= vuln 0) (member :FORTRESS collisions))) (= vuln *vulnerability*)) (update-slot 'miss))
    (if (< *vulnerability* vuln)  (update-slot 'hit))
    (setf *missiles* missiles)
    (setf *points* points)
    (setf *vulnerability* vuln))))

(defun record-features-track (data)
    (setf *rectangle1* (getf data :rectangle1))
    (setf *rectangle2* (getf data :rectangle2))
    (game-state-features-track data)
  (let ((points (getf data :pnts)) aim
        (ship (getf data :ship))
        (keys (getf data :keys)))
    (cond ((member :thrust keys) (setf *thrust* t))
          ((not (null keys)) (setf *turn* t))
          ((and *turn* *thrust* ship) (setf *turn* nil *thrust* nil) 
           (setf aim (chunk-slot-value-fct (buffer-read 'game-state) 'offaim))
           (if (< aim 40) (update-slot 'good-aim) (update-slot 'bad-aim))))
    (if (and (numberp points) (numberp *points*))
      (cond ((> points *points*) (update-slot  'rectangle))
            ((< points *points*) (update-slot  'death))))
      (setf *points* points)))

(defun outward (x y vx vy)
       (let* ((x1 (- x 355))
              (y1 (- y 315))
              (dist1 (sqrt (+ (* x1 x1) (* y1 y1) )))
              (x2 (- (+ x vx) 355))
              (y2 (- (- y vy) 315))
              (dist2 (sqrt (+ (* x2 x2) (* y2 y2)))))
         (- dist2 dist1)))

;;; This notes relevant information for the metacognitive module after every tick
(defun game-state-features-fortress (data) 
(setf *data* data)
 (if (car (no-output (buffer-chunk game-state)))
   (cond ((equal (getf data :screen-type) "score" )
          (mod-buffer-chunk 'game-state (list 'x nil 'y nil 'vx nil 'vy nil 'fortress-alive  nil 'ship-alive nil 'VULNERABILITY nil 'outward nil
                         'prop-to-big nil                     
                    'fortress-angle nil 'orientation nil   'thrust-angle nil 'speed nil 'angle nil  'time-to-outer nil 'status 'game-over)))
         ((getf (getf data :ship) :alive)
          (let* ((ship (getf data :ship))
                 (fortress-alive (if (getf (getf data :fortress) :alive) 'yes 'no))
                 (old-fortress-alive (chunk-slot-value-fct (buffer-read 'game-state) 'fortress-alive))
                 (x (getf ship :x))
                 (y (getf ship :y))
                 (angle (getf ship :angle))
                 (orientation (getf ship :orientation))
                 (vel (getf ship :speed))
                 (fortress-angle (mod (- (angle x (- 630 y) 355 315) 6) 360))
                 (vdir (getf ship :vdir))
                 (vx (getf ship :vx))
                 (vy (getf ship :vy))
                 (dsmall (dist-to-hex x y :radius 40))
                 (dbig (dist-to-hex x y :radius 200))
                 (vuln (getf data :vlner))
                 (outward (outward x y vx vy))
                 (bighex (aif (getf data :bighex) it 200))
                 (game-change (if (not (equal fortress-alive old-fortress-alive)) 'state-changed 
                                (chunk-slot-value-fct (buffer-read 'game-state) 'status)))
                 (time-to-outer  (if  (> outward 0) (travel-time-to-hex vel x y vx vy :radius bighex) 10))
                 (prop-to-big (if  (> outward 0) (/ dsmall (+ dsmall dbig)) 0))
                 )
            (mod-buffer-chunk 'game-state (list 'x x 'y y 'vx vx 'vy vy 'fortress-alive  fortress-alive 'ship-alive 'yes 'VULNERABILITY vuln
                    'fortress-angle fortress-angle 'orientation orientation 'game 'spacefortress 'outward outward 'prop-to-big prop-to-big
                     'thrust-angle (- vdir angle) 'speed vel 'angle angle 'time-to-outer time-to-outer 'status game-change))))
         (t 
          (mod-buffer-chunk 'game-state (list 'x nil 'y nil 'vx nil 'vy nil 'fortress-alive  (if (getf (getf data :fortress) :alive) 'yes 'no)
                                               'fortress-angle nil 'orientation nil 'VULNERABILITY (getf data :vlner) 'outward nil
                                               'prop-to-big nil
                                              'thrust-angle nil 'speed nil 'angle nil 'ship-alive 'no  'time-to-outer nil))))))

(defun game-state-features-track (data)
(setf *data* data)
 (if (car (no-output (buffer-chunk game-state))) 
   (cond ((equal (getf data :screen-type) "score" )
          (mod-buffer-chunk 'game-state (list 'status 'game-over 
                                              'orientation nil  'future-angle nil
                                                'moving-angle nil 'ship-alive nil 'r1 nil 'r2 nil
                                                'x nil 'y nil 'vx nil 'vy nil 'offaim nil
                                                'current-angle nil 'ship-speed nil)))
         ((getf (getf data :ship) :alive)
          (let* ((ship (getf data :ship))
                 (x (getf ship :x))
                 (y (getf ship :y))
                 (vx (getf ship :vx))
                 (vy (getf ship :vy))
                 (moving-angle (angle 0 0 vx vy))
                 (speed (getf ship :speed))
                 (orientation (mod (getf ship :orientation) 360))
                 (r1 (getf data :rectangle1))
                 (r2  (getf data :rectangle2))
                 (r1-angle  (box-angle r1))
                 (r2-angle  (box-angle r2))
                 (alive  (if (getf ship :alive) 'yes 'no))
                 (rect  (or (chunk-slot-value-fct (buffer-read 'goal) 'rectangle) r1)))
            (mod-buffer-chunk 'game-state (list 'orientation orientation 'game 'spacetrack
                                                'moving-angle moving-angle 'ship-alive alive 'r1 r1 'r2 r2
                                                'current-angle (if (equal rect r1) r1-angle r2-angle)
                                                'future-angle (if (equal rect r1) r2-angle)
                                                'x x 'y y 'vx vx 'vy vy 'offaim  (HIT-SIDES rect X Y VX VY)
                                               'ship-speed speed))))
         (t
          (mod-buffer-chunk 'game-state (list  'orientation nil  
                                                'moving-angle nil 'ship-alive 'no 'r1 nil 'r2 nil
                                                'x nil 'y nil 'vx nil 'vy nil 'offaim nil 'future-angle nil
                                                'current-angle nil 'ship-speed nil))))))


(defun update-slot (slot)   
       (schedule-mod-buffer-chunk 'imaginal (list slot 1) 0))

(defun speed-sf (vx vy) (sqrt (+ (* vx vx) (* vy vy)))) 

;calculates the difference between 2 angles dealing with looping at 360
(defun angle-offset (ang1 ang2)
          (let* ((offset (mod (- ang1 ang2) 360)))
            (if (> offset 180) (- 360 offset) offset)))

;what is angle of vector from first to second point?
(defun angle (x1 y1 x2 y2)
  (let* ((run (- x2 x1))
         (rise (- y2 y1))
         (a (if (not (zerop run))  (* (atan (/ rise run)) (/ 180 pi)))))
    (cond ((zerop run) (if (>= rise 0) 90 270))
          ((> run 0) (if (>= rise 0) a (+ 360 a)))
          (t (if (>= rise 0) (+ 180 a) (+ 180 a))))))


;how far from end will ship hit side of rectangle?
(defun hit-sides (rect x y vx vy)
  (max (hit-side (first rect) (second rect) x y vx vy) (hit-side (fourth rect) (third rect) x y vx vy)))

(defun hit-side (p1 p2 x y vx vy)
  (let* ((point (intersection-point p1 p2 x y vx vy)))
    (if  (and point (if (> vx 0) (< x (first point)) (> x (first point))))
      (sqrt (+ (sqr (- (first point) (second p2)))  (sqr (- (second point) (fourth p2))))) 0)))  

(defun safe-div (a b) (if (not (zerop b)) (/ a b) (progn (print 'bad-division) 1000)))

(defun intersection-point (p1 p2 x y vx vy) 
  (let* ((slope1  (safe-div (- vy) vx))
         (inter1 (- y (* x slope1)))
         (x1 (second p1))
         (x2 (second p2))
         (y1 (fourth p1))
         (y2 (fourth p2))
         (slope2 (if (not (equal x1 x2))  (safe-div (- y2 y1) (- x2 x1))))
         (inter2 (if slope2 (- y1 (* x1 slope2))))
         (point (line-intersect (list inter1 slope1)(if slope2 (list inter2 slope2) x2))))
    (if (and point (within (first point) x1  x2) (within (second point) y1  y2) ) point)))

(defun line-intersect (line1 line2)
  (cond ((and (listp line1) (listp line2))
         (let* ((a1 (first line1))
                (b1 (second line1))
                (a2 (first line2))
                (b2 (second line2)))
           (cond ((not (= b1 b2))
                  (let* ((x (safe-div (- a2 a1) (- b1 b2)))
                         (y (+ a1 (* b1 x))))
                    (list x y))))))
        ((listp line1)
         (list line2 (+ (first line1) (* line2 (second line1)))))
        ((listp line2)
         (list line1 (+ (first line2) (* line1 (second line2)))))))

(defun within (a b c)
  (or (and (>= b a) (>= a c)) (and (<= b a) (<= a c))))

;what is the orientation of a rectangle?
(defun box-angle (box)
  (if box
  (let* ((x1 (second (first box)))
         (x2 (second (second box)))
         (y1 (- (fourth (first box))))
         (y2 (- (fourth (second box)))))
          (angle x1 y1 x2 y2)) 0))         


;what angle to orient to to make a turn?   
(defun turn-angle (angle moving-angle) 
  (let* ((ntarget (if (> moving-angle angle) (+ 360 angle (- moving-angle)) (- angle moving-angle)))
         (naim (if (> ntarget 360) (- 270 (safe-div (- 360 ntarget) 2)) (+ 90 (/ ntarget 2)))))
    (mod (+ moving-angle naim) 360.0)))


;How far is the ship into the next rectangle?
(defun distance-in (rect x y vx vy)
  (let* ((p1 (first rect))
         (p2 (second rect))
         (p3 (third rect))
         (p4 (fourth rect))
         (result (list (in-help p1 p2 x y vx vy) (in-help p2 p3 x y vx vy)
                         (in-help p3 p4 x y vx vy) (in-help p4 p1 x y vx vy))))
    (if result (apply 'max result) -1000)))

(defun in-help (p1 p2 x y vx vy)
  (let* ((point (intersection-point p1 p2 x y vx vy))
         (distance (if point (sqrt (+ (sqr (- x (first point))) (sqr (- y (second point))))))))
    (cond ((null point) -1000)
          ((> vx 0) 
           (if (> x (first point)) distance (- distance)))
          (t (if (> x (first point)) (- distance) distance)))))

(defun sqr (x) (* x x))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Variable for the server name, defaults to localhost.

(defvar *server-name* "127.0.0.1")
(defvar *server-type* 'sfapi-tcp-client)

(defun set-server (server)
  (cond
    ((null server) #| do nothing |#)
    ((stringp server)
     (setf *server-name* server
           *server-type* 'sfapi-tcp-client))
    ((eq server :cluster-external-process)
     (set-server-script-location-for-cluster)
     (setf *server-type* 'sfapi-external-process-client))
    (t (error "Unknown server parameter: ~S" server))))

(defparameter *base-pathname* (make-pathname :name nil
                                             :type nil
                                             :defaults *load-pathname*))

(defun set-server-script-location-for-cluster (&optional (basedir *base-pathname*))
  "This function is hardcoded to expect a node binary in
<basedir>/node/bin/node and the server in
<basedir>/js-server/cli_server.js."
  #+sfapi
  (let* ((script (make-pathname :name "cli_server"
                                :type "js"
                                :directory (append (pathname-directory basedir) (list "js-server"))
                                :defaults basedir))
         (node (make-pathname :name "node"
                              :type nil
                              :directory (append (pathname-directory basedir) (list "node" "bin"))
                              :defaults basedir)))
    (setf *sfapi-server-command* (list (namestring node) (namestring script))))
  #-sfapi nil)


     ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;; Code to run conditions


 (defparameter *alpha* .2)
 (defparameter *factor* 1)
(defparameter *strategy* 1)


(defun run-experiment1 (n params flag &optional server)
  (setf *alpha* (first params))
  (setf *factor* (second params))
  (setf session1 nil)
  (setf *strategy* (fourth params))
  (run-n-learning n flag 1000 (third params) server))

(defun run-experiment2 (n params flag &optional server)
  (setf *alpha* (first params))
  (setf *factor* (second params))
  (setf session1 nil)
  (setf *strategy* (fourth params))
  (if (< (third params) 3)
      (run-n-learning n flag 1000 (third params) server)
    (run-n-transfer (round n 2) flag 1000 (third params) server)))

(defun run-n-transfer (n flag speed game &optional server)
  (let (hold
        (first (case *strategy*
           (1  youturn-instructions1)
           (2 youturn-instructions2)
           (3 youturn-instructions3)
           (4 youturn-instructions4))))
  (setf *instructions* (union first spacetrack-instructions :test 'equal))
  (cond ((equal game 3)
         (setf *data-hook* 'record-features-fortress) 
         (play-sf-games 1 :speed speed :draw flag :condition "youturn")
         (do ((i 1 (1+ i))
              (quads (list (quad-lists 1)) (cons (quad-lists (1+ i)) quads)))
             ((= i n) (setf hold (reverse quads)))
           (play-sf-games 1 :speed speed :draw flag :cont t))
         (setf session1 (set-difference (no-output (pp)) (union session1 starting)))
         (sgp :sf-data-hook record-features-track) 
         (play-sf-games 1 :speed speed :draw flag :condition "nav" :cont t)
         (do ((i 1 (1+ i))
              (quads (list (quad-lists 1)) (cons (quad-lists (1+ i)) quads)))
             ((= i n) (append hold (reverse quads)))
           (play-sf-games 1 :speed speed :draw flag :cont t)))
        ((equal game 4)
         (setf *data-hook* 'record-features-track) 
         (play-sf-games 1 :speed speed :draw flag :condition "nav")
         (do ((i 1 (1+ i))
              (quads (list (quad-lists 1)) (cons (quad-lists (1+ i)) quads)))
             ((= i n) (setf hold (reverse quads)))
           (play-sf-games 1 :speed speed :draw flag :cont t))
         (setf session1 (set-difference (no-output (pp)) (union session1 starting)))
         (sgp :sf-data-hook  record-features-fortress) 
         (play-sf-games 1 :speed speed :draw flag :condition "youturn" :cont t)
         (do ((i 1 (1+ i))
              (quads (list (quad-lists 1)) (cons (quad-lists (1+ i)) quads)))
             ((= i n) (append hold (reverse quads)))
           (play-sf-games 1 :speed speed :draw flag :cont t))))))


(defun run-n-learning (n flag speed game &optional server)
  (set-server server)
  (cond ((equal game 0)
         (setf *data-hook* 'record-features-fortress) 
         (case *strategy*
           (1 (setf *instructions* youturn-instructions1))
           (2 (setf *instructions* youturn-instructions2))
           (3 (setf *instructions* youturn-instructions3))
           (4 (setf *instructions* youturn-instructions4))
           )
         (play-sf-games 1 :speed speed :draw flag :condition "youturn"))
        ((equal game 1)
         (setf *data-hook* 'record-features-fortress) 
         (case *strategy*
           (1 (setf *instructions* autoturn-instructions1))
           (2 (setf *instructions* autoturn-instructions2))
           (3 (setf *instructions* autoturn-instructions3))
           (4 (setf *instructions* autoturn-instructions4))
           )
         (play-sf-games 1 :speed speed :draw flag :condition "autoturn"))
        (t (setf *data-hook* 'record-features-track) 
           (setf *instructions* spacetrack-instructions)
           (play-sf-games 1 :speed speed :draw flag :condition "nav")))
  (do ((i 1 (1+ i))
       (quads (list (quad-lists 1)) (cons (quad-lists (1+ i)) quads)))
      ((= i n) (reverse quads))
    (play-sf-games 1 :speed speed :draw flag :cont t)))



(defun play-autoturn (n &optional (speed 1)(draw t))
  (setf *strategy* 1)
  (play-sf-games n :speed speed :draw draw :condition "autoturn" :cont t))

(defun play-youturn (n &optional (speed 1)(draw t))
  (setf *strategy* 1)
  (play-sf-games n :speed speed :draw draw :condition "youturn" :cont t))

(defun play-navigation (n &optional (speed 1)(draw t))
  (setf *data-hook* 'record-features-track) 
  (setf *instructions* spacetrack-instructions)
  (play-sf-games n :speed speed :draw draw :condition "nav" :cont t))


(defun quad-lists (i)
   (cons (session-report)
          (mapcar #'(lambda (x) (append (list i (* 1.0 (tracker-temperature x))) (tracker-params x)))   (trackers (get-module :tracker)))))

(defun session-report ()
          (let* ((set2 (set-difference (no-output (pp)) (union session1 starting)))
                (temp0 (mapcar 'p-count starting))
                (temp1 (if session1 (mapcar 'p-count session1) (mapcar 'p-count set2)))
                (temp2 (if session1  (mapcar 'p-count set2) (mapcar 'p-count session1)))
                (types (list (- (length starting) (count 0 temp0)) (- (length temp1) (count 0 temp1)) (- (length temp2) (count 0 temp2))))
                (tokens (list (apply '+ temp0) (apply '+ temp1) (apply '+ temp2))))
            (reset-p-count (get-module :p-count))
             (append (list (length starting) (if session1 (length session1) (length set2)) (if session1 (length set2) (length session1))) types tokens)))

(defun default-chunks (lis) (eval (cons 'add-dm (mapcar #'(lambda (x) (list x 'isa 'chunk)) lis))))

(defparameter *data-hook* 'record-features-fortress)


(defparameter starting '(GET-STARTED GET-STARTED-AGAIN GET-STARTED-NEW-GAME START-PLAYING-SPACETRACK START-PLAYING-SPACEFORTRESS START-PLAYING-AGAIN-SPACETRACK 
                                     START-PLAYING-AGAIN-SPACEFORTRESS DETECT-TRIAL-END DETECT-DESTROYED DO-SOMETHING DO-SOMETHING-FAST OBJECT-PRESENT OBJECT-NOT-PRESENT TEST-DIMENSION-SUCCESS 
                                     TEST-DIMENSION-FAIL TURN-TO-POINT1 SKIP-TURN START-PRESS-CLOCKWISE PRESS-CLOCKWISE-TAP START-PRESS-COUNTERCLOCKWISE PRESS-COUNTERCLOCKWISE-TAP TAP-KEY TAP-W TAPPPING-D
                                     PRESSING-D FINISH-PRESS-D PRESSING-A FINISH-PRESS-A TAPPING-A CALCULATE-TURN-ANGLE THRUST-NEW-RECTANGLE FINISH-THRUST INCREMENT-ANGLE DECREMENT-ANGLE EMERGENCY-THRUST 
                                     DELAY-OK PRESS-SPACEBAR DOUBLE-SPACEBAR DOUBLE-SHOT))

(defparameter session1 nil)

(defparameter autoturn-instructions1
  '((starter
  ISA OPERATOR
   PRE  SPACEFORTRESS
   action test-object
   arg1 fortress-alive
   success fortress
   fail nofortress)
(spacefortress
   ISA OPERATOR
   PRE  fortress
   action  test-dimension 
   arg1 time-to-outer 
   success delay?
   fail tapthrust)
(spacefortress1
   ISA OPERATOR
   PRE  nofortress
   action  test-dimension 
   arg1 time-to-outer 
   success delay?
   fail tapthrust)
(delay
  ISA OPERATOR
   PRE  delay?
   action  check-delay
   success vulnerability?)
(vulnerability
  ISA OPERATOR
   PRE  vulnerability?
   ACTION  test-dimension
   arg1 vulnerability
  success kill-it
   fail shoot)
(kill-it 
      isa operator
      pre kill-it
      action double-spacebar
      post done)
(shoot 
      isa operator
      pre shoot
      action spacebar
      post done)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))

(defparameter autoturn-instructions2
  '((starter
  ISA OPERATOR
   PRE  SPACEFORTRESS
   action test-object
   arg1 fortress-alive
   success fortress
   fail nofortress)
(spacefortress
   ISA OPERATOR
   PRE  fortress
   action  test-dimension 
   arg1 prop-to-big 
   success tapthrust
   fail delay?)
(spacefortress1
   ISA OPERATOR
   PRE  nofortress
   action  test-dimension 
   arg1 prop-to-big 
   success tapthrust
   fail delay?)
(delay
  ISA OPERATOR
   PRE  delay?
   action  check-delay
   success vulnerability?)
(vulnerability
  ISA OPERATOR
   PRE  vulnerability?
   ACTION  test-dimension
   arg1 vulnerability
  success kill-it
   fail shoot)
(kill-it 
      isa operator
      pre kill-it
      action double-spacebar
      post done)
(shoot 
      isa operator
      pre shoot
      action spacebar
      post done)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))


(defparameter autoturn-instructions3
  '((starter
  ISA OPERATOR
   PRE  SPACEFORTRESS
   action test-object
   arg1 fortress-alive
   success fortress
   fail nofortress)
(spacefortress
   ISA OPERATOR
   PRE  fortress
   action  test-dimension 
   arg1 time-to-outer 
   success delay?
   fail tapthrust)
(spacefortress1
   ISA OPERATOR
   PRE  nofortress
   action  test-dimension
   arg1 outward 
   success tapthrust
   fail delay?)
(delay
  ISA OPERATOR
   PRE  delay?
   action  check-delay
   success vulnerability?)
(vulnerability
  ISA OPERATOR
   PRE  vulnerability?
   ACTION  test-dimension
   arg1 vulnerability
  success kill-it
   fail shoot)
(kill-it 
      isa operator
      pre kill-it
      action double-spacebar
      post done)
(shoot 
      isa operator
      pre shoot
      action spacebar
      post done)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))

(defparameter autoturn-instructions4
  '((starter
  ISA OPERATOR
   PRE  SPACEFORTRESS
   action test-object
   arg1 fortress-alive
   success fortress
   fail nofortress)
(spacefortress
   ISA OPERATOR
   PRE  fortress
   action  test-dimension 
   arg1 prop-to-big 
   success tapthrust
   fail delay?)
(spacefortress1
   ISA OPERATOR
   PRE  nofortress
   action  test-dimension
   arg1 outward 
   success tapthrust
   fail delay?)
(delay
  ISA OPERATOR
   PRE  delay?
   action  check-delay
   success vulnerability?)
(vulnerability
  ISA OPERATOR
   PRE  vulnerability?
   ACTION  test-dimension
   arg1 vulnerability
  success kill-it
   fail shoot)
(kill-it 
      isa operator
      pre kill-it
      action double-spacebar
      post done)
(shoot 
      isa operator
      pre shoot
      action spacebar
      post done)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))

(defparameter youturn-instructions1
  '((starter
  ISA OPERATOR
   PRE  SPACEFORTRESS
   action test-object
   arg1 fortress-alive
   success fortress
   fail nofortress)
  (spacefortress
   ISA OPERATOR
   PRE  fortress
   action  test-dimension 
   arg1 time-to-outer 
   success shooting
   fail thrusting)
(spacefortress1
   ISA OPERATOR
   PRE  nofortress
   action  test-dimension 
   arg1 time-to-outer 
   success shooting
   fail thrusting)
(aim
  ISA OPERATOR
   PRE  shooting
   action  turn-to-point
   arg1 fortress-angle
   post delay?)
(delay
  ISA OPERATOR
   PRE  delay?
   action  check-delay
   success vulnerability?)
(vulnerability
  ISA OPERATOR
   PRE  vulnerability?
   ACTION  test-dimension
   arg1 vulnerability
  success kill-it
   fail shoot)
(kill-it 
      isa operator
      pre kill-it
      action double-spacebar
      post done)
(shoot 
      isa operator
      pre shoot
      action spacebar
      post done)
(judge-speed
  ISA OPERATOR
   PRE  thrusting
   ACTION  test-dimension 
   arg1 speed
   fail speed-up
   success slow-down)
(speed-up
  ISA OPERATOR
   PRE  speed-up
   ACTION  increment-angle
   arg1 thrust-angle
   post tapthrust)
(slow-down
  ISA OPERATOR
   PRE  slow-down
   ACTION  decrement-angle
   arg1 thrust-angle
   post tapthrust)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))

(defparameter youturn-instructions2
  '((starter
  ISA OPERATOR
   PRE  SPACEFORTRESS
   action test-object
   arg1 fortress-alive
   success fortress
   fail nofortress)
(spacefortress
   ISA OPERATOR
   PRE  fortress
   action  test-dimension 
   arg1 prop-to-big 
   success thrusting
   fail shooting)
(spacefortress1
   ISA OPERATOR
   PRE  nofortress
   action  test-dimension 
   arg1 prop-to-big 
   success thrusting
   fail shooting)
(aim
  ISA OPERATOR
   PRE  shooting
   action  turn-to-point
   arg1 fortress-angle
   post delay?)
(delay
  ISA OPERATOR
   PRE  delay?
   action  check-delay
   success vulnerability?)
(vulnerability
  ISA OPERATOR
   PRE  vulnerability?
   ACTION  test-dimension
   arg1 vulnerability
  success kill-it
   fail shoot)
(kill-it 
      isa operator
      pre kill-it
      action double-spacebar
      post done)
(shoot 
      isa operator
      pre shoot
      action spacebar
      post done)
(judge-speed
  ISA OPERATOR
   PRE  thrusting
   ACTION  test-dimension 
   arg1 speed
   fail speed-up
   success slow-down)
(speed-up
  ISA OPERATOR
   PRE  speed-up
   ACTION  increment-angle
   arg1 thrust-angle
   post tapthrust)
(slow-down
  ISA OPERATOR
   PRE  slow-down
   ACTION  decrement-angle
   arg1 thrust-angle
   post tapthrust)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))


(defparameter youturn-instructions3
  '((starter
  ISA OPERATOR
   PRE  SPACEFORTRESS
   action test-object
   arg1 fortress-alive
   success fortress
   fail nofortress)
  (spacefortress
   ISA OPERATOR
   PRE  fortress
   action  test-dimension 
   arg1 time-to-outer 
   success shooting
   fail thrusting)
(spacefortress1
   ISA OPERATOR
   PRE  nofortress
   action  test-dimension
   arg1 outward 
   success thrusting
   fail shooting)
(aim
  ISA OPERATOR
   PRE  shooting
   action  turn-to-point
   arg1 fortress-angle
   post delay?)
(delay
  ISA OPERATOR
   PRE  delay?
   action  check-delay
   success vulnerability?)
(vulnerability
  ISA OPERATOR
   PRE  vulnerability?
   ACTION  test-dimension
   arg1 vulnerability
  success kill-it
   fail shoot)
(kill-it 
      isa operator
      pre kill-it
      action double-spacebar
      post done)
(shoot 
      isa operator
      pre shoot
      action spacebar
      post done)
(judge-speed
  ISA OPERATOR
   PRE  thrusting
   ACTION  test-dimension 
   arg1 speed
   fail speed-up
   success slow-down)
(speed-up
  ISA OPERATOR
   PRE  speed-up
   ACTION  increment-angle
   arg1 thrust-angle
   post tapthrust)
(slow-down
  ISA OPERATOR
   PRE  slow-down
   ACTION  decrement-angle
   arg1 thrust-angle
   post tapthrust)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))

(defparameter youturn-instructions4
  '((starter
  ISA OPERATOR
   PRE  SPACEFORTRESS
   action test-object
   arg1 fortress-alive
   success fortress
   fail nofortress)
(spacefortress
   ISA OPERATOR
   PRE  fortress
   action  test-dimension 
   arg1 prop-to-big 
   success thrusting
   fail shooting)
(spacefortress1
   ISA OPERATOR
   PRE  nofortress
   action  test-dimension
   arg1 outward 
   success thrusting
   fail shooting)
(aim
  ISA OPERATOR
   PRE  shooting
   action  turn-to-point
   arg1 fortress-angle
   post delay?)
(delay
  ISA OPERATOR
   PRE  delay?
   action  check-delay
   success vulnerability?)
(vulnerability
  ISA OPERATOR
   PRE  vulnerability?
   ACTION  test-dimension
   arg1 vulnerability
  success kill-it
   fail shoot)
(kill-it 
      isa operator
      pre kill-it
      action double-spacebar
      post done)
(shoot 
      isa operator
      pre shoot
      action spacebar
      post done)
(judge-speed
  ISA OPERATOR
   PRE  thrusting
   ACTION  test-dimension 
   arg1 speed
   fail speed-up
   success slow-down)
(speed-up
  ISA OPERATOR
   PRE  speed-up
   ACTION  increment-angle
   arg1 thrust-angle
   post tapthrust)
(slow-down
  ISA OPERATOR
   PRE  slow-down
   ACTION  decrement-angle
   arg1 thrust-angle
   post tapthrust)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))

(defparameter spacetrack-instructions
'((spacetrack
  ISA OPERATOR
   PRE  spacetrack
   action  test-dimension 
   arg1 offaim
   success  adjust
   FAIL  test-speed)
(turn-to-target
  ISA OPERATOR
   PRE  adjust
   action calculate-turn-angle
   arg1 current-angle
   post tapthrust)
(speedtest
  ISA OPERATOR
   PRE  test-speed
   action TEST-DIMENSION
   arg1 ship-speed
   fail speedup
   success turn)
(dm speedup
      isa operator
      pre speedup
      action turn-to-point
      arg1 current-angle
      post tapthrust)
(turn
  ISA OPERATOR
   PRE  turn
   action calculate-turn-angle
   arg1 future-angle
   post maketurn2)
(thrust-turn
  ISA OPERATOR
   PRE  maketurn2
   action thrust-to-turn
   post done)
(thrustaction
      isa operator
      pre tapthrust
      action tap-key
      arg1 thrust
      post done)))




(defparameter *instructions* youturn-instructions1)

