"use strict";

(function (exports) {

var util = typeof require === 'undefined' ? window.util:require('./util');

var GAMETIME = 180000;

function Config() {
    Object.call(this);
    this.game_type = "none";

    this.screen_width = 710;
    this.screen_height = 626;
    this.game_time = 60000;

    this.destroy_fortress = 100;
    this.ship_death_penalty = 100;
    this.missile_penalty = 2;

    this.max_points = 3000;
    this.max_bonus = 60;

    this.auto_turn = false;

    this.fortress = {};
    this.fortress.sectorSize = 10;
    this.fortress.lockTime = 1000;
    this.fortress.vlnerTime = 250;
    this.fortress.vlnerThreshold = 10;
    this.fortress.collisionRadius = 18;
    this.fortress.explodeDuration = 1000;

    // Hexagons

    this.bigHex = 200;
    this.smallHex = 40;

    this.staircase = {};
    this.staircase.enabled = false;
    this.staircase.hexContraction = 5;
    this.staircase.hexExpansion = 15;
    this.staircase.minHexDistance = 20;

    this.ship = {};
    this.ship.startAngle = 0;
    this.ship.turnSpeed = 6;
    this.ship.acceleration = 0.3;
    this.ship.collisionRadius = 10;
    this.ship.explodeDuration = 1000;
    this.ship.startPosition = {x: 235, y: 315};
    this.ship.startVelocity = {x: Math.cos(util.deg2rad(60)), y: Math.sin(util.deg2rad(60))};

    this.missiles = {};
    this.missiles.collisionRadius = 5;
    this.missiles.speed = 20;

    this.shells = {};
    this.shells.collisionRadius = 3;
    this.shells.speed = 6;

    this.nav = {};
    this.nav.staircase = false;
    this.nav.staircase_delta = 15;
    this.nav.staircase_decrease_threshold = 1; //Decrease half_width if deaths exceed this
    this.nav.staircase_increase_threshold = 1; //Increase half_width if deaths exceed this
    this.nav.rect_width = 110;
    this.nav.enter_rectangle = 25;
    this.nav.min_angle = 30;
    this.nav.max_angle = 150;
    this.nav.min_rect_width = 50;

    this.flappy = false;
    this.flappy_nest = true;
    this.flappy_eggs = true;

    return this;
}

function YouturnConfig() {
    Config.call(this);
    this.game_type = "explode";
    this.game_time = GAMETIME;
    this.auto_turn = false;
    return this;
}

function FlappyYouturnConfig() {
    YouturnConfig.call(this);
    this.game_type = "flappy_youturn";
    this.flappy = true;
    this.max_bonus = 120;
    return this;
}

function AutoturnConfig() {
    Config.call(this);
    this.game_type = "autoturn";
    this.game_time = GAMETIME;
    this.auto_turn = true;
    return this;
}

function FlappyAutoturnConfig() {
    AutoturnConfig.call(this);
    this.game_type = "flappy_autoturn";
    this.flappy = true;
    return this;
}

function AutoturnPracticeConfig() {
    Config.call(this);
    this.game_type = "autoturn";
    this.auto_turn = true;
    return this;
}

function AutoturnBonusConfig() {
    Config.call(this);
    this.game_type = "staircase";
    this.staircase.enabled = true;
    this.max_points = 2000;
    this.max_bonus = 40;
    this.destroy_fortress = 200;
    this.ship_death_penalty = 200;
    this.missile_penalty = 4;

    return this;
}

function NavConfig() {
    Config.call(this);
    this.game_type = "nav";
    this.game_time = GAMETIME;
    this.max_points = 1000;
    this.max_bonus = 30;
    this.ship.startPosition = {x:350, y:350};
    this.ship.startVelocity = {x:1, y:0};
    return this;
}

function RaceConfig() {
    Config.call(this);
    this.game_type = "race";
    this.game_time = 120000;
    this.max_points = 1000;
    this.max_bonus = 60;
    this.ship.startPosition = {x:0, y:0};
    this.ship.startVelocity = {x:1, y:0};
    return this;
}

exports.NavConfig = NavConfig;
exports.AutoturnConfig = AutoturnConfig;
exports.FlappyAutoturnConfig = FlappyAutoturnConfig;
exports.YouturnConfig = YouturnConfig;
exports.FlappyYouturnConfig = FlappyYouturnConfig;

}) (typeof exports === 'undefined' ? this['config']={}:exports);
