var roundfn = Math.round

var DEBUG_ANGLE_AMT = 1;

var POLAR_A_SCALE = 2;
var POLAR_R_SCALE = 1.3;

var COLLIDE_FORTRESS = 1;
var COLLIDE_SHIP = 2;
var COLLIDE_BIGHEX = 3;
var COLLIDE_SMALLHEX = 4;
var COLLIDE_RECTANGLE = 5;

var TURN_CODES = [null, 'left', 'right'];

var KEY_CODES = ['fire',
                 'left',
                 'right',
                 'thrust'];

var EVENT_CODES = ['explode-smallhex',
                   'explode-bighex',
                   'ship-respawn',
                   'fortress-fired',
                   'shell-hit-ship',
                   'missile-fired',
                   'hit-fortress',
                   'fortress-destroyed',
                   'fortress-respawn',
                   'vlner-reset',
                   'vlner-increased',
                   'explode-rectangle',
                   'enter-seg2',
                   'leave-seg2',

                   'press-left',
                   'press-right',
                   'press-thrust',
                   'press-fire',
                   'release-left',
                   'release-right',
                   'release-thrust',
                   'release-fire'];

function keycode2name(ev) {
    if (ev.which == 32) {
        return 'fire';
    } else if (ev.which == 65) {
        return 'left';
    } else if (ev.which == 68) { // 69) {
        return 'right';
    } else if (ev.which == 87) { // 188) {
        return 'thrust';
    // } else if (ev.which == 81) {
    //     return 'quit';
    } else if (ev.which == 27) {
        return 'quit';
    } else {
        return null;
    }
}

function deg2rad(deg) {
    return deg * Math.PI / 180;
}

function rad2deg(rad) {
    return rad / Math.PI * 180;
}

function mod(x,y) {
    return ((x%y)+y)%y;
}

function Vector(x,y) {
    Object.call(this);
    this.x = x;
    this.y = y;
    return this;
}

function V(x,y) {
    return new Vector(x,y);
}

Vector.prototype = {};

Vector.prototype.add = function (v) {
    this.x += v.x;
    this.y += v.y;
    return this;
};

Vector.prototype.sub = function(v) {
  this.x -= v.x;
  this.y -= v.y;
  return this;
};

Vector.prototype.scalar_mult = function(alpha) {
  this.x *= alpha;
  this.y *= alpha;
  return this;
};

Vector.prototype.copy = function (v) {
    this.x = v.x;
    this.y = v.y;
};

Vector.prototype.duplicate = function() {
  return new Vector(this.x,this.y);
};

Vector.prototype.norm = function() {
  return Math.sqrt(Math.pow(this.x,2) + Math.pow(this.y,2));
};

Vector.prototype.normalize = function() {
  var norm = this.norm();
  if (norm > 0) {
    this.x /= norm;
    this.y /= norm;
  }
  return this;
};

Vector.prototype.eq = function(v) {
    return this.x === v.x && this.y === v.y;
};

function angle_diff(angle1, angle2) {
    // It finally works--don't touch it!
    if (angle1 < 90) {
        if (angle2 > angle1+180) return 360-angle2 + angle1;
        else return angle1 - angle2;
    } else if (angle1 < 180) {
        if (angle2 > angle1+180) return 360-angle2 + angle1;
        else return angle1 - angle2;
    } else if (angle1 < 270) {
        if (angle2 < angle1-180) return -(angle2 + 360-angle1);
        else return angle1 - angle2;
    } else {
        if (angle2 < angle1-180) return -(angle2 + 360-angle1);
        else return angle1 - angle2;
    }
}

function std_angle(a) {
    a = mod(a, 360);
    if (a >= 180) a -= 360;
    else if (a <= -180) a += 360;
    return a;
}

Vector.prototype.polar = function(relative) {
    // var v= V(mod(rad2deg(Math.atan2(315-this.y, 355-this.x)),360)-180,
    //          Math.sqrt(Math.pow(355-this.x,2) + Math.pow(315-this.y,2)));

    // var v= V((mod(rad2deg(Math.atan2(315-this.y, 355-this.x)),360)-180) * POLAR_A_SCALE,
    //          Math.sqrt(Math.pow(355-this.x,2) + Math.pow(315-this.y,2)) * POLAR_R_SCALE);

    // var v= V((rad2deg(Math.atan2(315-this.y, 355-this.x))-180) * POLAR_A_SCALE,
    //          Math.sqrt(Math.pow(355-this.x,2) + Math.pow(315-this.y,2)) * POLAR_R_SCALE);

    var a = rad2deg(Math.atan2(315-this.y, 355-this.x));
    var ra = rad2deg(Math.atan2(315-relative.y, 355-relative.x));

    var v= V(std_angle(angle_diff(a,ra)) * POLAR_A_SCALE,
             Math.sqrt(Math.pow(355-this.x,2) + Math.pow(315-this.y,2)) * POLAR_R_SCALE);

    return v;
}

function dotProduct (v1, v2) {
    return v1.x * v2.x + v1.y * v2.y;
}

function projection(a, b) {
    return dotProduct(a,b) / b.norm();
}

function angle_to(v1, v2) {
    var a = Math.atan2(-(v2.y-v1.y),
                       v2.x-v1.x);
    if (a < 0) { a += Math.PI*2; }
    return rad2deg(a);
}

function Rect (x, y, w, h) {
    this.x = x;
    this.y = y;
    this.w = w;
    this.h = h;

    return this;
}

Rect.prototype = {};

function Timer(time) {
    Object.call(this);
    this.ticks = 0;
    this.time = time || 0;
    this.last_tick = 0;
    return this;
}

Timer.prototype = {};

Timer.prototype.tick = function (ms) {
    this.time += ms;
    this.last_tick = ms;
    this.ticks += 1;
};

Timer.prototype.reset = function () {
    this.time = 0;
    this.ticks = 0;
};

Timer.prototype.elapsed = function () {
    return this.time;
};

function KeyState() {
    Object.call(this);

    this.events = [];
    this.keys = {'left': false,
                 'right': false,
                 'thrust': false,
                 'fire': false};
    return this;
}

KeyState.prototype = {};

KeyState.prototype.press = function (which) {
    // ignore repeat events
    if (!this.keys[which]) this.events.push({'type': 1, 'key': which});
    this.keys[which] = true;
};

KeyState.prototype.release = function (which) {
    this.keys[which] = false;
    this.events.push({'type': 0, 'key': which});
};

KeyState.prototype.clear = function () {
    this.events = [];
};

function Wireframe(points, runs, color) {
    Object.call(this);
    this.points = points;
    this.runs = runs;
    this.color = color;
    return this;
}

Wireframe.prototype = {};

// Wireframe.prototype.rotate_translate = function (pair, x, y, angle) {
//     return [ pair[0] * Math.cos(angle) - pair[1] * Math.sin(angle) + x,
//             -pair[1] * Math.sin(angle) + pair[1] * Math.cos(angle) + y];
// }

// Wireframe.prototype.translate_run = function (run, x, y, angle) {
//     var i, new_run = [];
//     for (i=0; i<run.length; i++) {
//         new_run[i] = this.rotate_translate(run[i][0]);
//     }
// };

Wireframe.prototype.draw = function (canvas, x, y, angle) {
    var r, i;
    canvas.save();
    canvas.strokeStyle = this.color;
    canvas.translate(x,y);
    canvas.scale(1,-1);
    canvas.rotate(deg2rad(angle));
    for (r=0; r<this.runs.length; r++) {
        canvas.beginPath();
        canvas.moveTo(this.points[this.runs[r][0]][0], this.points[this.runs[r][0]][1]);
        for (i=1; i<this.runs[r].length; i++) {
            canvas.lineTo(this.points[this.runs[r][i]][0], this.points[this.runs[r][i]][1]);
        }
        canvas.stroke();
    }
    canvas.restore();
};

var missileWireframe = new Wireframe([[0, 0], [-25, 0], [-5, 5], [-5, -5]],
                                     [[0,1], [0, 2], [0, 3]],
                                    '#FFFFFF');
var shellWireframe = new Wireframe([[-8, 0], [0, -6], [16, 0], [0, 6], [-8, 0]],
                                   [[0,1,2,3,0]],
                                   '#FF0000');
var shipWireframe = new Wireframe([[-18,0], [18,0], [0,0], [-18,18], [-18,-18]],
                                  [[0,1], [3, 2, 4]],
                                  '#FFFF00');

var birdWireframe = new Wireframe([[0,-18], [0,18], [0,0], [18,18]],
                                  [[0,1], [3, 2]],
                                  '#FFFF00');

var fortressWireframe = new Wireframe([[0, 0], [36, 0], [18, -18], [0,-18], [18,18], [0,18]],
                                      [[0, 1], [3,2,4,5]],
                                     '#FFFF00');

function Entity() {
    Object.call(this);

    this.position = V(0,0);
    this.velocity = V(0,0);
    this.angle = 0;
    this.collision_radius = 0;
    this.alive = true;

    return this;
}

Entity.prototype = {};

function distance(v1, v2) {
    return Math.sqrt(Math.pow(v1.x - v2.x, 2) + Math.pow(v1.y - v2.y, 2));
}

function collided(e1, e2) {
    var d = distance(e1.position, e2.position);
    return d <= e1.collision_radius + e2.collision_radius;
}

function outside_game_area(e) {
    return e.position.x < 0 || e.position.x > 710 || e.position.y < 0 || e.position.y > 626;
}

function drawExplosion(canvas, x, y, frame) {
    if (frame >= 0 && frame < 10) {
        var img = g_images['explosion-animation'];
        canvas.save();
        canvas.translate(x,y);
        canvas.scale(1,-1);
        canvas.drawImage(img, frame * 96, 0, 96, 96,
                         -96/2,-96/2,96,96);
        canvas.restore();
    }
}

function drawWireframeExplosion(canvas, x, y) {
    var ofs = 0;
    var radius;
    for (radius=15; radius<70; radius+=8) {
        var angle;
        ofs += 3;
        if (radius < 60) { canvas.strokeStyle = '#FFFF00'; }
        else { canvas.strokeStyle = '#FF0000'; }
        for (angle=0; angle<360; angle += 30) {
            canvas.beginPath();
            canvas.arc(x, y, radius, deg2rad(angle+ofs),deg2rad(angle+ofs+10),false);
            canvas.stroke();
        }
    }
    canvas.strokeStyle = '#FFFF00';
    canvas.beginPath();
    canvas.arc(x, y, 7, 0, Math.PI*2, false);
    canvas.stroke();
}

function Ship(config) {
    Entity.call(this);
    this.startAngle = 0;
    this.angle = this.startAngle;
    this.startVelocity = V(config.shipStartVelocity.x, config.shipStartVelocity.y);
    this.startPosition = V(config.shipStartPosition.x, config.shipStartPosition.y);
    this.turn_speed = 6;
    this.acceleration = 0.3;
    this.collision_radius = 10;

    this.position.copy(this.startPosition);
    this.velocity.copy(this.startVelocity);
    this.turn_flag = null;
    this.thrust_flag = false;
    this.deathTimer = new Timer();
    return this;
}

Ship.prototype = new Entity();

Ship.prototype.reset = function () {
    this.alive = true;
    var a = Math.atan2(this.position.y-315, this.position.x-355);
    var d = distance (this.startPosition, {x:355,y:315});
    this.position.x = 355 + d * Math.cos(a);
    this.position.y = 315 + d * Math.sin(a);
    // console.log(this.position, rad2deg(a));
    this.velocity.x = Math.cos(a+Math.PI/2);
    this.velocity.y = -Math.sin(a+Math.PI/2);
    this.angle = angle_to(this.position, {x:355,y:315});
};

Ship.prototype.draw = function (canvas) {
    if (this.alive) {
        shipWireframe.draw(canvas, this.position.x, this.position.y, this.angle);

        // if (this.thrust_flag)
        //     draw_image(canvas, 'ship-thrust', this.position.x, this.position.y, this.angle);
        // else
        //     draw_image(canvas, 'ship', this.position.x, this.position.y, this.angle);
    } else {
        var frame = Math.floor(this.deathTimer.ticks / 3);
        drawWireframeExplosion(canvas, this.position.x, this.position.y, frame);
    }
};

Ship.prototype.drawPolar = function (canvas) {
    var p = this.position.polar();
    // if (this.alive) {
    // canvas.strokeStyle = "#00FF00";
    canvas.fillStyle = "#FFFF00";

    shipWireframe.draw(canvas, p.x, p.y,
                       angle_to(this.position, {x:355,y:315})-this.angle+90);


    // for (var a=p.x-20.5; a<p.x+20; a++) {
    //     for (var r=p.y-20; r<p.y+20; r++) {
    //         var cp = V(Math.cos(deg2rad(a)) * r+355, Math.sin(deg2rad(a)) * r+315);
    //         var d = distance(this.position, cp);
    //         if(d <= this.collision_radius) {
    //             // canvas.fillStyle = 'rgb('+(d*25.5)+','+(d*25.5)+',0)';
    //             canvas.fillRect(a, r, 1.5, 1.5);
    //         }
    //     }
    // }



    // canvas.beginPath();
    // canvas.arc (p.x, p.y, this.collision_radius, 0, Math.PI*2);
    // canvas.fill();
    // canvas.stroke();

    // } else {
    //     var frame = Math.floor(this.deathTimer.ticks / 3);
    //     drawWireframeExplosion(canvas, p.x, p.y, frame);
    // }
};

Ship.prototype.drawSprite = function (canvas, gp, use_angle) {
    var p = this.position.polar(this.position);
    canvas.save();
    canvas.translate(roundfn(gp.x), roundfn(gp.y + p.y));
    if (use_angle) {
        canvas.rotate(deg2rad(this.angle-angle_to(this.position, V(355,315))));
    } else {
        if (this.alive && this.thrust_flag) canvas.rotate(deg2rad(-8));
    }
    canvas.drawImage(g_images['bird'], roundfn(-g_images['bird'].width/2), roundfn(-g_images['bird'].height/2));
    canvas.restore();
}

function Fortress(config) {
    Entity.call(this);
    this.collision_radius = 18;
    this.sector_size = config.fortress_sector_size;
    this.lockTime = config.fortress_lock_time;
    this.position.x = 355;
    this.position.y = 315;

    this.timer = new Timer();
    this.vulnerabilityTimer = new Timer(config.vlner_time);
    this.deathTimer = new Timer();

    return this;
}

function draw_image(canvas, img, x, y, angle) {
    canvas.save();
    canvas.translate(x,y);
    canvas.scale(1,-1);
    canvas.rotate(deg2rad(angle+90));
    canvas.drawImage(g_images[img], -g_images[img].width/2, -g_images[img].height/2);
    canvas.restore();
}

Fortress.prototype = new Entity();

Fortress.prototype.draw = function (canvas) {
    if (this.alive) {
        fortressWireframe.draw(canvas, this.position.x, this.position.y, this.angle);
        // draw_image(canvas, 'fortress', this.position.x, this.position.y, this.angle);
    } else {
        var frame = Math.floor(this.deathTimer.ticks / 3);
        drawWireframeExplosion(canvas, this.position.x, this.position.y, frame);
    }
};

Fortress.prototype.drawPolar = function (ship_pos, canvas) {
    sp = ship_pos.polar();
    p = V(0,0);
    if (this.alive) {
        fortressWireframe.draw(canvas, p.x+sp.x, p.y, 270);
        // draw_image(canvas, 'fortress', this.position.x, this.position.y, this.angle);
    }//  else {
    //     var frame = Math.floor(this.deathTimer.ticks / 3);
    //     drawWireframeExplosion(canvas, p.x+sp.x, p.y, frame);
    // }
}

Fortress.prototype.drawSprite = function (canvas, relative, gp, neggs) {
    sp = relative.polar(relative);
    p = V(0,0);
    if (this.alive) {
        var x=0, y=0;
        if (neggs<=4) { x=0; y=0; }
        else if (neggs<=7) { x=5; y=5; }
        else if (neggs<=9) { x=10; y=10; }
        else { x=15; y=15; }
        for (var i=neggs; i>0; i--) {
            canvas.drawImage(g_images['egg'],
                             gp.x+p.x+sp.x-g_images['nest'].width/2 + x,
                             gp.y+p.y-y);
            x += 10;
            if (i==5) { x=0; y=0; }
            if (i==8) { x=5; y=5; }
            if (i==10) { x=10; y=10; }
            if (i>10) { x=15; y=15; }
        }
        canvas.drawImage(g_images['nest'], gp.x+p.x+sp.x-g_images['nest'].width/2, gp.y+p.y);
    }
}

function Missile(x, y, angle) {
    Entity.call(this);

    this.collision_radius = 5;
    this.speed = 20;

    this.position.x = x;
    this.position.y = y;
    this.angle = angle;
    this.velocity.x = this.speed * Math.cos(deg2rad(angle));
    this.velocity.y = -this.speed * Math.sin(deg2rad(angle));
    return this;
}

Missile.prototype = new Entity();

Missile.prototype.draw = function (canvas) {
    // draw_image(canvas, 'missile', this.position.x, this.position.y, this.angle);
    missileWireframe.draw(canvas, this.position.x, this.position.y, this.angle);
};

Missile.prototype.drawPolar = function (canvas) {
    // draw_image(canvas, 'missile', this.position.x, this.position.y, this.angle);
    var p = this.position.polar();
    missileWireframe.draw(canvas, p.x, p.y, angle_to(this.position, {x:355,y:315})-this.angle+90);
};

Missile.prototype.drawSprite = function (canvas, relative, gp) {
    // draw_image(canvas, 'missile', this.position.x, this.position.y, this.angle);
    var p = this.position.polar(relative);
    // missileWireframe.draw(canvas, p.x, p.y, angle_to(this.position, {x:355,y:315})-this.angle+90);
    canvas.drawImage(g_images['egg'], roundfn(gp.x+p.x), roundfn(gp.y+p.y));
};


function Shell(config, x, y, angle) {
    Entity.call(this);

    this.collision_radius = 3;
    this.speed = config.shell_speed;
    this.acceleration = 0.1;

    this.position.x = x;
    this.position.y = y;
    this.angle = angle;
    this.velocity.x = this.speed * Math.cos(deg2rad(angle));
    this.velocity.y = -this.speed * Math.sin(deg2rad(angle));
    return this;
}

Shell.prototype = new Entity();

Shell.prototype.draw = function (canvas) {
    // draw_image(canvas, 'shell', this.position.x, this.position.y, this.angle);
    shellWireframe.draw(canvas, this.position.x, this.position.y, this.angle);
};

Shell.prototype.drawPolar = function (ctx) {
    var p = this.position.polar();
    var opening = 20;
    var stalk = opening - 3;
    // shellWireframe.draw(canvas, p.x, p.y, this.angle-angle_to(this.position, {x:355,y:315})+90);
    ctx.save();
    ctx.strokeStyle = "#FF0000";
    ctx.strokeStyle = "#CC0000";
    ctx.strokeRect(p.x-stalk, 0, stalk+this.collision_radius-3, p.y-this.collision_radius);
    ctx.strokeStyle = "#FF0000";
    ctx.strokeRect(p.x-opening, p.y-this.collision_radius, opening+this.collision_radius, this.collision_radius*2);

    // ctx.beginPath();
    // ctx.arc (p.x, p.y, this.collision_radius, 0, Math.PI*2);
    // ctx.fill();
    // ctx.stroke();

    ctx.restore();
};

Shell.prototype.drawSprite = function (ctx, relative, gp) {
    var p = this.position.polar(relative);
    p.x -= 15;
    var h = p.y-30;
    ctx.drawImage( g_images['tube-top'], roundfn(gp.x+p.x), roundfn(gp.y+p.y) );
    // ctx.save();
    // ctx.translate(p.x+3,p.y-4);
    // ctx.translate(roundfn(gp.x+p.x+3),roundfn(gp.y+p.y));
    // console.log('shell', p.x);
    var i = g_images['tube-middle'];
    ctx.drawImage( i, 0, 0, i.width, i.height, roundfn(gp.x+p.x+3), roundfn(gp.y+p.y-h), i.width, roundfn(h) );
    // ctx.restore();
};

function Hexagon (radius, img) {
    Object.call(this);
    var x1 = Math.floor(355-radius);
    var x2 = Math.floor(355-radius*0.5);
    var x3 = Math.floor(355+radius*0.5);
    var x4 = Math.floor(355+radius);
    var y1 = 315;
    var y2 = Math.floor(315-radius*Math.sin(Math.PI*2/3));
    var y3 = Math.floor(315+radius*Math.sin(Math.PI*2/3));
    this.points = [V(x1,y1),
                   V(x2,y2),
                   V(x3,y2),
                   V(x4,y1),
                   V(x3,y3),
                   V(x2,y3),
                   V(x1,y1)];
    this.radius = radius;
    var s = distance(this.points[0], this.points[1]);
    this.apothem = s / 2 * Math.tan(Math.PI * (6-2) / (2 * 6));
    this.img = img;
    return this;
}

Hexagon.prototype = {};

Hexagon.prototype.inside = function (v) {
    var i;
    for (i=0; i<this.points.length-1; i++) {
        var nx, ny, dx, dy;
        nx = -(this.points[i+1].y - this.points[i].y);
        ny =   this.points[i+1].x - this.points[i].x;
        dx = v.x - this.points[i].x;
        dy = v.y - this.points[i].y;
        if (nx * dx + ny * dy < 0) {
            return false;
        }
    }
    return true;
};

Hexagon.prototype.draw = function (canvas) {
    var i;
    canvas.beginPath();
    canvas.strokeStyle = '#00FF00';
    canvas.moveTo(this.points[0].x, this.points[0].y);
    for (i=1; i<this.points.length; i++) {
        canvas.lineTo(this.points[i].x, this.points[i].y);
    }
    canvas.closePath();
    canvas.stroke();
};

Hexagon.prototype.scanPolar = function (ctx) {
    ctx.fillStyle = '#00FF00';
    for(var r = this.apothem; r<this.radius; r++) {
        for(var a = 0; a<360; a++) {
            var p = V(Math.cos(deg2rad(a)) * r+355, Math.sin(deg2rad(a)) * r+315);
            if(!this.inside(p)) ctx.fillRect(a-180, r, 1, 1);
        }
    }
}

function drawEllipse(ctx, x, y, w, h) {
  var kappa = .5522848,
      ox = (w / 2) * kappa, // control point offset horizontal
      oy = (h / 2) * kappa, // control point offset vertical
      xe = x + w,           // x-end
      ye = y + h,           // y-end
      xm = x + w / 2,       // x-middle
      ym = y + h / 2;       // y-middle

  ctx.beginPath();
  ctx.moveTo(x, ym);
  ctx.bezierCurveTo(x, ym - oy, xm - ox, y, xm, y);
  ctx.bezierCurveTo(xm + ox, y, xe, ym - oy, xe, ym);
  // ctx.bezierCurveTo(xe, ym + oy, xm + ox, ye, xm, ye);
  // ctx.bezierCurveTo(xm - ox, ye, x, ym + oy, x, ym);

    //ctx.closePath(); // not used correctly, see comments (use to close off open path)

    ctx.stroke();
}

Hexagon.prototype.drawPolar = function (canvas) {
    var i;
    var eh = this.radius-this.apothem;
    var ew = Math.ceil(this.points[1].polar().x-this.points[0].polar().x);

    // console.log('hex', eh*3*POLAR_R_SCALE, ew);

    // canvas.beginPath();
    // canvas.strokeStyle = '#00FF00';
    var p = this.points[0].polar();

    // canvas.moveTo(p.x, p.y);
    for (i=1; i<this.points.length; i++) {
        var np = this.points[i].polar();
        drawEllipse(canvas, (p.x+1), (np.y-eh*POLAR_R_SCALE),
                    ew, eh*3 * POLAR_R_SCALE);
        p = np;
    }
    // canvas.stroke();
};

Hexagon.prototype.drawSprite = function (canvas) {
    var eh = this.radius-this.apothem;
    var ew = Math.ceil(this.points[1].polar().x-this.points[0].polar().x);
    var ofs = Math.floor(eh*POLAR_R_SCALE);

    var p = this.points[0].polar();
    p.x = Math.round(p.x);
    p.y = Math.round(p.y);
    for(var i=-1; i<7; i++) {
        canvas.drawImage(this.img, i * (this.img.width - 1), p.y-ofs);
    }
};

function Game(config, gnum) {
    Object.call(this);

    this.screen_id = 'game';
    this.config = config;
    this.game_number = gnum;

    this.init_basic_game_state();

    this.ship = new Ship(config);
    this.fortress = new Fortress(config);
    this.shells = [];
    this.missiles = [];
    this.score.vulnerability = 0;

    return this;
}

Game.prototype = {};

Game.prototype.init_basic_game_state = function () {
    this.collisions = [];
    this.score = {};
    this.score.pnts = 0;
    this.score.raw_pnts = 0;

    this.keyState = new KeyState();

    this.gameTimer = new Timer();
    this.currentTick = 0;

    this.currentFrameEvents = [];

    this.shipDeaths = 0;
};

Game.prototype.addDebugShells = function () {
    for (var s=0; s<360; s += 10) {
        var p = V(Math.cos(deg2rad(s)) * 100 + 355,
                  Math.sin(deg2rad(s)) * 100 + 315);
        this.shells.push(new Shell(this.config, p.x, p.y, 0));
    }
}

Game.prototype.init = function () {
    $("#experiment_area").html('<div class="canvasarea"><canvas id="screen-canvas" width="710px" height="600px"></canvas></div>');

    this.canvas = document.getElementById('screen-canvas').getContext('2d');
    this.timer = window.setInterval($.proxy(this.tick, this), 33);
    this.addEventListeners();

    this.createBackground();

    if (!this.config.staircase || !exp.bigHex || !exp.smallHex) {
        exp.bigHex = this.config.bigHex;
        exp.smallHex = this.config.smallHex;
    }
    this.bighex = new Hexagon(exp.bigHex, g_images['hills']);
    this.smallhex = new Hexagon(exp.smallHex, g_images['ceiling']);

    this.canvas.lineWidth = 1.4;
    this.last_tick_time = null;
    this.tinc = 0;

    // Debug stuff
    this.angle = 0;
    // this.addDebugShells();

    this.fireMissileSound = sounds['fire-missile'];// $('#fire_missile')
    this.fireShellSound = sounds['fire-shell']; // $('#fire_shell')
    this.vlnerResetSound = sounds['vlner-reset']; // $('#vlner_reset')
    this.explosionSound = sounds['explosion'];

    exp.resetGameData();

    // Record game specific data
    exp.gameLog.push({'type': this.config.game_type, 'version': 1, 'bighex': this.bighex.radius, 'smallhex': this.smallhex.radius});

    exp.lg('start', {'n': this.game_number, 'bighex': this.bighex.radius, 'smallhex': this.smallhex.radius});
};

Game.prototype.addEventListeners = function () {
    $(document).on('keydown', $.proxy(this.keydown_event, this));
    $(document).on('keyup', $.proxy(this.keyup_event, this));
};

Game.prototype.clearEvents = function () {
    $(document).off('keydown', $.proxy(this.keydown_event));
    $(document).off('keyup', $.proxy(this.keyup_event));
};

Game.prototype.play = function (x) {
    // x.prop('currentTime', 0);
    // x.trigger('play');
    x.play();
};

Game.prototype.keydown_event = function(ev) {
    var sym = keycode2name(ev);
    if (sym == 'quit' && isDebugMode()) {
        ev.preventDefault();
        window.clearInterval(this.timer);
        this.clearEvents();
    } else if (sym !== null) {
        this.keyState.press(sym);
        ev.preventDefault();
    }
};

Game.prototype.keyup_event = function (ev) {
    var sym = keycode2name(ev);
    if (sym !== null) {
        this.keyState.release(sym);
        ev.preventDefault();
    }
};

Game.prototype.getWorldState = function () {
    var missiles = [], shells = [];
    var i;

    for (i=0; i<this.missiles.length; i++) {
        missiles.push([Math.round(this.missiles[i].position.x*1000), Math.round(this.missiles[i].position.y*1000), Math.round(this.missiles[i].angle)
                      ]);
    }

    for (i=0; i<this.shells.length; i++) {
        shells.push([Math.round(this.shells[i].position.x*1000), Math.round(this.shells[i].position.y*1000), Math.round(this.shells[i].angle)
                    ]);
    }

    var line = [this.gameTimer.elapsed(),
                // ship
                this.ship.alive?1:0,
                Math.round(this.ship.position.x*1000),
                Math.round(this.ship.position.y*1000),
                // We want 2 decimals of precision
                Math.round(this.ship.velocity.x * 1000),
                Math.round(this.ship.velocity.y * 1000),
                Math.floor(this.ship.angle),
                // fortress
                this.fortress.alive?1:0,
                Math.floor(this.fortress.angle),
                // missiles
                missiles,
                // shells
                shells,
                // score
                this.score.pnts,
                this.score.vulnerability,
                // key state
                this.ship.thrust_flag?1:0,
                TURN_CODES.indexOf(this.ship.turn_flag)
                // this.keyState.keys['thrust']?1:0,
                // this.keyState.keys['left']?1:0,
                // this.keyState.keys['right']?1:0,
                // this.keyState.keys['fire']?1:0,
                ];

    if (this.currentFrameEvents.length > 0) {
        line.push(this.currentFrameEvents);
        this.currentFrameEvents = [];
    }

    return line;
};

Game.prototype.recordEverything = function () {
    exp.gameLog.push(this.getWorldState());
};

Game.prototype.recordMinimal = function () {
    this.recordFrameEvents();
    if (exp.gameKeyLog[exp.gameKeyLog.length-1][0] == this.currentTick) {
        this.recordWorldState(true);
    }
};

Game.prototype.addEvent = function (tag, extra) {
    // console.log(tag,extra);
    var idx = EVENT_CODES.indexOf(tag);
    var val;
    if (idx >= 0) val = idx; else val = tag;
    if (extra) {
        this.currentFrameEvents.push([val, extra]);
    } else {
        this.currentFrameEvents.push(val);
    }
};

Game.prototype.penalize = function (amt) {
    this.score.raw_pnts -= amt;
    this.score.pnts -= amt;
    if (this.score.pnts < 0) this.score.pnts = 0;
};

Game.prototype.reward = function (amt) {
    this.score.raw_pnts += amt;
    this.score.pnts += amt;
};

Game.prototype.monitor_ship_respawn = function () {
    if (this.ship.alive === false && this.ship.deathTimer.elapsed() >= 1000) {
        this.ship.reset();
        this.fortress.timer.reset();
        if (this.config.polar) {
            for(var i=0; i<this.shells.length; i++) {
                if (this.shells[i].keep_around) {
                    this.shells[i].alive = false;
                    this.shells.splice(i, 1);
                }
            }
        }
        this.addEvent('ship-respawn');
    }
};


Game.prototype.update_fortress = function () {
    var dx = this.ship.position.x - this.fortress.position.x;
    var dy = -(this.ship.position.y - this.fortress.position.y);
    var angle_to_ship = mod(rad2deg(Math.atan2(dy, dx)), 360);

    if (!this.fortress.alive && this.fortress.deathTimer.elapsed() > 1000) {
        this.fortress.alive = true;
        this.addEvent('fortress-respawn');
    }

    if (this.ship.alive) {
        this.fortress.angle = mod(Math.floor(angle_to_ship / this.fortress.sector_size) * this.fortress.sector_size, 360);
    }
    if (this.fortress.angle !== this.fortress.last_angle) {
        this.fortress.last_angle = this.fortress.angle;
        this.fortress.timer.reset();
    }
    if (this.fortress.timer.elapsed() >= this.fortress.lockTime && this.ship.alive && this.fortress.alive) {
        this.shells.push(new Shell(this.config, this.fortress.position.x, this.fortress.position.y, angle_to_ship));
        this.fortress.timer.reset();
        this.play(this.fireShellSound);
        this.addEvent('fortress-fired');
    }
};

Game.prototype.process_key_state = function () {
    var e;
    for (e=0; e<this.keyState.events.length; e++) {
        var evt = this.keyState.events[e];
        this.addEvent((evt['type']===1 ? 'press':'release') + '-' + evt['key']);
        this.process_key_event(evt);
    }
    this.keyState.clear();
};

Game.prototype.process_key_event = function (evt) {
    if (evt['type'] == 1) {
        if (evt['key'] == 'left' || evt['key'] == 'right') {
            this.ship.turn_flag = evt['key'];
        } else if (evt['key'] == 'thrust') {
            this.ship.thrust_flag = true;
        } else if (evt['key'] == 'fire') {
            this.fire_missile();
        }
    } else {
        if (evt['key'] == 'left' || evt['key'] == 'right') {
            this.ship.turn_flag = null;
        } else if (evt['key'] == 'thrust') {
            this.ship.thrust_flag = false;
        }
    }
};

Game.prototype.fire_missile = function () {
    if (this.ship.alive) {
        var m = new Missile(this.ship.position.x, this.ship.position.y, this.ship.angle);
        this.missiles.push(m);
        this.penalize(this.config.missile_penalty);
        this.addEvent('missile-fired');
    }
};

Game.prototype.kill_ship = function () {
    if (this.ship.alive) {
        this.penalize(this.config.ship_death_penalty);
        this.ship.alive = false;
        this.shipDeaths += 1;
        this.ship.deathTimer.reset();
        this.play(this.explosionSound);
    }
};

Game.prototype.move_ship = function () {
    if (this.config.auto_turn) {
        this.ship.angle = Math.floor(angle_to(this.ship.position, this.fortress.position));
        //console.log(this.ship.position.x, this.ship.position.y, this.fortress.position.x, this.fortress.position.y, angle_to(this.ship.position, this.fortress.position));
    } else {
        var delta = this.ship.turn_speed;
        if (this.config.polar) delta *= -1;
        if (this.ship.turn_flag == 'right') {
            this.ship.angle = mod(this.ship.angle - delta, 360);
        } else if (this.ship.turn_flag == 'left') {
            this.ship.angle = mod(this.ship.angle + delta, 360);
        }
    }
    if (this.ship.thrust_flag) {
        this.ship.velocity.x += this.ship.acceleration * Math.cos(deg2rad(this.ship.angle));
        this.ship.velocity.y += this.ship.acceleration * Math.sin(deg2rad(this.ship.angle));
    }
    // Max Speed
    // if (this.ship.velocity.x > this.config.ship_max_vel) {
    //     this.ship.velocity.x = this.config.ship_max_vel;
    //     console.log('max x');
    // } else if (this.ship.velocity.x < -this.config.ship_max_vel) {
    //     this.ship.velocity.x = -this.config.ship_max_vel;
    //     console.log('max -x');
    // }
    // if (this.ship.velocity.y > this.config.ship_max_vel) {
    //     this.ship.velocity.y = this.config.ship_max_vel;
    //     console.log('max y');
    // } else if (this.ship.velocity.y < -this.config.ship_max_vel) {
    //     this.ship.velocity.y = -this.config.ship_max_vel;
    //     console.log('max -y');
    // }
    this.ship.position.x += this.ship.velocity.x;
    this.ship.position.y -= this.ship.velocity.y;
};

Game.prototype.update_ship = function () {
    if (this.ship.alive) {
        this.move_ship();
        if (!this.bighex.inside(this.ship.position)) {
            this.kill_ship();
            this.collisions.push(COLLIDE_BIGHEX);
            this.addEvent('explode-bighex');
        } else if (this.smallhex.inside(this.ship.position)) {
            this.kill_ship();
            this.collisions.push(COLLIDE_SMALLHEX);
            this.addEvent('explode-smallhex');
        }
    }
};

Game.prototype.update_missiles = function () {
    var i;
    for (i=0; i<this.missiles.length; i++) {
        this.missiles[i].position.add(this.missiles[i].velocity);
        if (collided (this.missiles[i], this.fortress)) {
            this.missiles[i].alive = false;
            this.collisions.push(COLLIDE_FORTRESS);
            this.addEvent('hit-fortress');
            if (this.fortress.alive) {
                if (this.fortress.vulnerabilityTimer.elapsed() >= this.config.vlner_time) {
                    this.score.vulnerability += 1;
                    this.addEvent('vlner-increased');
                    this.play(this.fireMissileSound);
                } else {
                    if (this.score.vulnerability >= this.config.vlner_threshold + 1) {
                        this.fortress.alive = false;
                        this.fortress.deathTimer.reset();
                        this.reward(this.config.destroy_fortress);
                        this.play(this.fireMissileSound);
                        this.play(this.explosionSound);
                        this.addEvent('fortress-destroyed');
                    } else {
                        this.play(this.vlnerResetSound);
                        this.addEvent('vlner-reset');
                    }
                    this.score.vulnerability = 0;
                }
                this.fortress.vulnerabilityTimer.reset();
            }
        } else if (outside_game_area(this.missiles[i])) {
            this.missiles[i].alive = false;
        }
    }
    var alive = [];
    for (i=0; i<this.missiles.length; i++) {
        if (this.missiles[i].alive) {
            alive.push(this.missiles[i]);
        }
    }
    this.missiles = alive;
};

Game.prototype.update_shells = function () {
    var i;
    for (i=0; i<this.shells.length; i++) {
        this.shells[i].position.add(this.shells[i].velocity);
        if (this.ship.alive && collided (this.shells[i], this.ship)) {
            this.collisions.push(COLLIDE_SHIP);
            if (this.config.polar) {
                this.shells[i].keep_around = true;
                this.shells[i].velocity.x = 0;
                this.shells[i].velocity.y = 0;
            } else {
                this.shells[i].alive = false;
            }
            this.kill_ship();
            this.addEvent('shell-hit-ship');
        } else if (outside_game_area(this.shells[i])) {
            this.shells[i].alive = false;
        }
    }
    var alive = [];
    for (i=0; i<this.shells.length; i++) {
        if (this.shells[i].alive) {
            alive.push(this.shells[i]);
        }
    }
    this.shells = alive;
};

Game.prototype.draw_score = function (pairs) {
    var label_width = 89;
    var label_height = 32;
    var pad = 16;
    var score_y = 520.5;
    var start = (710-89*pairs.length)/2;
    var i;

    // this.canvas.clearRect(start, score_y, label_width*pairs.length, label_height*2);
    this.canvas.drawImage(g_images['scoreboard'], (710-90*2)/2, 520);

    // this.canvas.beginPath();
    // this.canvas.strokeStyle = "#00FF00";
    // this.canvas.rect(start + 0.5, score_y, label_width*pairs.length, label_height*2);
    // this.canvas.moveTo(start+label_width+0.5, score_y);
    // this.canvas.lineTo(start+label_width+0.5, score_y+label_height*2);
    // this.canvas.moveTo(start+0.5, score_y+label_height);
    // this.canvas.lineTo(start+0.5+label_width*pairs.length, score_y+label_height);
    // this.canvas.stroke();

    this.canvas.textAlign = 'center';
    this.canvas.textBaseline = 'middle';

    // this.canvas.fillStyle = "#00FF00";
    this.canvas.fillStyle = "#FFFFFF";
    this.canvas.shadowColor = "#000000"
    this.canvas.shadowOffsetX = 2;
    this.canvas.shadowOffsetY = 2;
    this.canvas.font = "14px sans-serif";
    for (i=0; i<pairs.length; i++) {
        this.canvas.fillText(pairs[i].title, start + label_width*i + label_width/2, score_y+18);
    }

    this.canvas.font = "16px sans-serif";
    this.canvas.fillStyle = "#FFFF00";
    for (i=0; i<pairs.length; i++) {
        this.canvas.fillText(pairs[i].value.toString(), start + label_width*i + label_width/2, score_y+label_height+15);
    }
};

Game.prototype.draw = function () {
    var i;

    this.canvas.clearRect(0, 0, 710, 626);
    // this.canvas.drawImage(g_images['bg'], 0, 0);

    this.canvas.save();
    this.canvas.translate(0,-50);

    this.bighex.draw(this.canvas);
    this.smallhex.draw(this.canvas);
    this.ship.draw(this.canvas);
    this.fortress.draw(this.canvas);
    for (i=0; i<this.shells.length; i++) {
        if (!collided(this.shells[i], this.fortress)) {
            this.shells[i].draw(this.canvas);
        }
    }
    for (i=0; i<this.missiles.length; i++) {
        this.missiles[i].draw(this.canvas);
    }

    this.canvas.restore();

    this.draw_score([{title: 'PNTS', value: this.score.pnts},
                     {title: 'VLNER', value: this.score.vulnerability}]);
};

Game.prototype.drawPolar = function () {
    var i;

    this.canvas.clearRect(0, 0, 710, 626);
    // this.canvas.drawImage(g_images['bg'], 0, 0);

    this.canvas.save();
    this.canvas.translate(355,200);
    // this.canvas.translate(0,-70);
    // this.canvas.scale(1.5,1.5);

    var ship_polar = this.ship.position.polar();
    this.canvas.translate(-ship_polar.x, 0);

    if (this.fortress.alive) this.canvas.strokeStyle = '#00FF00';
    else this.canvas.strokeStyle = '#999999';

    // this.bighex.scanPolar(this.canvas);
    this.bighex.drawPolar(this.canvas);
    this.smallhex.drawPolar(this.canvas);

    // Draw it a second time to simulate scrolling
    this.canvas.save();

    this.canvas.translate(360*POLAR_A_SCALE,0);
    this.bighex.drawPolar(this.canvas);
    this.smallhex.drawPolar(this.canvas);
    this.canvas.translate(-360*2*POLAR_A_SCALE,0);
    this.bighex.drawPolar(this.canvas);
    this.smallhex.drawPolar(this.canvas);

    this.canvas.restore();

    this.ship.drawPolar(this.canvas);
    if (this.config.polar_fortress_visible)
        this.fortress.drawPolar(this.ship.position, this.canvas);
    for (i=0; i<this.shells.length; i++) {
        if (!collided(this.shells[i], this.fortress)) {
            this.shells[i].drawPolar(this.canvas);
        }
    }
    if (this.config.polar_missile_visible) {
        for (i=0; i<this.missiles.length; i++) {
            this.missiles[i].drawPolar(this.canvas);
        }
    }
    this.canvas.restore();

    this.draw_score([{title: 'PNTS', value: this.score.pnts},
                     {title: 'COUNT', value: this.score.vulnerability}]);
};

Game.prototype.drawBricks = function (ctx, img, y) {
    var w = img.width;
    for (var i=0; i<6; i++) {
        ctx.drawImage(img, i*120, y);
    }
}

Game.prototype.createBackground = function () {
    this.bg = document.createElement('canvas');
    this.bg.width = g_images['hills'].width*6;
    this.bg.height = 300;
    var ctx = this.bg.getContext('2d');

    ctx.drawImage(g_images['bg'], 0, 0, g_images['bg'].width, g_images['bg'].height,
                  0, 26,
                  this.bg.width, g_images['bg'].height-40);
    this.drawBricks(ctx, g_images['bricks'], 26);
    this.drawBricks(ctx, g_images['bricks'], 277);
    // Hexagons
    for(var i=0, j=0;
        i<this.bg.width;
        i += g_images['hills'].width,
        j += g_images['ceiling'].width) {
        ctx.drawImage(g_images['hills'], i, 226);
        ctx.drawImage(g_images['ceiling'], j, 46);
    }
    this.bgData = ctx.getImageData(0, 0, this.bg.width, this.bg.height);

    this.drawBricks(ctx, g_images['darkbricks'], 26);
    this.drawBricks(ctx, g_images['darkbricks'], 277);
    this.darkData = ctx.getImageData(0, 0, this.bg.width, this.bg.height);
}

Game.prototype.drawAngleMarks = function () {
    var p = V(0,0);
    this.canvas.textAlign = 'center';
    this.canvas.textBaseline = 'middle';
    this.canvas.font = "9px sans-serif";
    this.canvas.beginPath();
    for (var a=0; a<360; a+= 10) {
        p.x = Math.cos(deg2rad(a)) * 20 + 355;
        p.y = Math.sin(deg2rad(a)) * 20 + 315;
        var p2 = p.polar(this.ship.position);
        this.canvas.moveTo(p2.x, 0);
        this.canvas.lineTo(p2.x, 400);
        this.canvas.fillText(a.toString(), p2.x, 100);
    }
    this.canvas.strokeStyle = "#FF0000";
    this.canvas.stroke();
}

Game.prototype.drawSprite = function () {
    var i;
    var ship_polar = V(235,315).polar(this.ship.position);
    var gp = V(0, 0);
    // ship_polar.x = Math.round(ship_polar.x);

    this.canvas.clearRect(0, 0, 710, 626);

    // console.log(ship_polar.x);
    // this.canvas.putImageData(this.bgData, mod(ship_polar.x, this.bgData.width), 140);
    var data;
    if (this.fortress.alive) data = this.bgData;
    else data = this.darkData;

        this.canvas.putImageData(data, roundfn(ship_polar.x), 140);
        if (ship_polar.x > 0)
            this.canvas.putImageData(data, roundfn(ship_polar.x-this.bgData.width), 140);
        else
            this.canvas.putImageData(data, roundfn(ship_polar.x+this.bgData.width), 140);

    this.canvas.save();
    // this.canvas.translate(0,-60);
    gp.y -= 60;

    // this.canvas.drawImage(g_images['bg'], 0, 0, g_images['bg'].width, g_images['bg'].height,
    //                       -mod(ship_polar.x, g_images['bg'].width), 226,
    //                       710*3, g_images['bg'].height-40);


    // this.canvas.save();
    // this.canvas.translate(-mod(Math.floor(ship_polar.x), g_images['bricks'].width), 200);

    // this.drawBricks(277);
    // this.drawBricks(26);
    // this.canvas.restore();

    // this.canvas.save();
    // this.canvas.translate(-mod(ship_polar.x, this.bighex.img.width), 200);


    // this.bighex.drawSprite(this.canvas);
    // this.smallhex.drawSprite(this.canvas);

    // this.canvas.restore();

    // this.canvas.translate(355,200);
    gp.x += 355;
    gp.y += 200;


    // this.canvas.translate(0,-70);
    // this.canvas.scale(1.5,1.5);

    // this.canvas.translate(-ship_polar.x, 0);
    // this.canvas.save();
    // this.canvas.drawImage(g_images['bg'], 0, 0, g_images['bg'].width, g_images['bg'].height,
    //                       -710, 26, 710*3, g_images['bg'].height-40);
    // this.canvas.restore();

    // this.drawAngleMarks();

    if (this.fortress.alive) this.canvas.strokeStyle = '#00FF00';
    else this.canvas.strokeStyle = '#999999';


    // this.bighex.scanPolar(this.canvas);
    // this.smallhex.drawPolar(this.canvas);

    // Draw it a second time to simulate scrolling
    // this.canvas.save();

    // this.canvas.translate(360*POLAR_A_SCALE,0);
    // this.bighex.drawSprite(this.canvas);
    // this.smallhex.drawSprite(this.canvas);
    // this.canvas.translate(-360*2*POLAR_A_SCALE,0);
    // this.bighex.drawSprite(this.canvas);
    // this.smallhex.drawSprite(this.canvas);

    // this.canvas.restore();

    this.ship.drawSprite(this.canvas, gp, !this.config.auto_turn);
    // this.canvas.translate(ship_polar.x, 0);

    if (this.config.polar_fortress_visible)
        this.fortress.drawSprite(this.canvas, this.ship.position, gp, this.score.vulnerability);
    for (i=0; i<this.shells.length; i++) {
        if (!collided(this.shells[i], this.fortress)) {
            this.shells[i].drawSprite(this.canvas, this.ship.position, gp);
        }
    }
    if (this.config.polar_missile_visible) {
        for (i=0; i<this.missiles.length; i++) {
            this.missiles[i].drawSprite(this.canvas, this.ship.position, gp);
        }
    }
    this.canvas.restore();

    this.canvas.save();
    this.canvas.translate(0,-60);

    this.draw_score([{title: 'POINTS', value: this.score.pnts},
                     {title: 'COUNT', value: this.score.vulnerability}]);
    this.canvas.restore();
};

Game.prototype.step_timers = function () {
    this.currentTick += 1;
    this.fortress.timer.tick(this.tinc);
    this.fortress.deathTimer.tick(this.tinc);
    this.fortress.vulnerabilityTimer.tick(this.tinc);
    this.ship.deathTimer.tick(this.tinc);
};

Game.prototype.update_time = function () {
    var now = getCurrentTime();
    if (this.last_tick_time === null) { this.last_tick_time = now; }
    this.tinc = now - this.last_tick_time;
    this.last_tick_time = now;
    // Advance the game timer but not the other timers because that's
    // how original autoturn works.
    this.gameTimer.tick(this.tinc);
};

Game.prototype.step_one_tick = function () {
    this.update_time();
    this.process_key_state();
    this.monitor_ship_respawn();
    this.update_ship();
    this.update_fortress();
    this.update_shells();
    this.update_missiles();

    // console.log('alive', this.gameTimer.elapsed(), this.tinc, this.ship.deathTimer.elapsed(), this.ship.alive);

    this.recordEverything();
    this.step_timers();
};

Game.prototype.debug_one_tick = function () {
    this.process_key_state();
    var mult = 1;
    if (this.ship.thrust_flag) mult = 0.1;
    if (this.ship.turn_flag == 'left') {
        this.angle += DEBUG_ANGLE_AMT * mult;
    } else if (this.ship.turn_flag == 'right') {
        this.angle -= DEBUG_ANGLE_AMT * mult;
    }
    if (this.angle < 0) this.angle += 360;
    if (this.angle >= 360) this.angle -= 360;
    this.ship.position.x = Math.cos(deg2rad(this.angle)) * 40 + 355;
    this.ship.position.y = Math.sin(deg2rad(this.angle)) * 40 + 315;
};

Game.prototype.tick = function () {
    // console.log('tick', getCurrentTime());

    this.step_one_tick();
    // this.debug_one_tick();

    if (this.config.polar) this.drawSprite();
    else this.draw();

    if (this.gameTimer.elapsed() >= this.config.game_time) {
        exp.nextScreen();
    }
};

Game.prototype.calculate_bonus = function () {
    this.bonus = Math.ceil((this.score.pnts / this.config.max_points) * this.config.max_bonus);
};

Game.prototype.cleanup_preamble = function () {
    window.clearInterval(this.timer);
    this.clearEvents();
    this.calculate_bonus();
    exp.gameReward = this.bonus;
    exp.gamePoints = this.score.pnts;
    exp.addReward(this.bonus);
};

Game.prototype.cleanup_main = function () {
    if (this.config.staircase) {
        // Update hex sizes for next game
        if (this.shipDeaths <= 1) {
            exp.bigHex -= this.config.hexContraction;
            exp.smallHex += this.config.hexContraction;
            // Don't make it too small.
            if (exp.bigHex - exp.smallHex < this.config.minHexDistance) {
                exp.bigHex += this.config.hexContraction;
                exp.smallHex -= this.config.hexContraction;
            }
        } else {
            exp.bigHex += this.config.hexExpansion;
            exp.smallHex -= this.config.hexExpansion;
            // don't exceed original size
            if (exp.bigHex > this.config.bigHex) {
                exp.bigHex = this.config.bigHex;
            }
            if (exp.smallHex < this.config.smallHex) {
                exp.smallHex = this.config.smallHex;
            }
        }
    }
};

Game.prototype.cleanup_finish = function () {
    // Send it all to the server
    stats = {'deaths': this.shipDeaths, 'bonus': this.bonus, 'points': this.score.pnts, 'rawpnts': this.score.raw_pnts};
    exp.gameLog.push(stats);
    exp.com.sendGameData(this.game_number, exp.gameLog);
    exp.lg('end', stats);
};

Game.prototype.cleanup = function () {
    this.cleanup_preamble();
    this.cleanup_main();
    this.cleanup_finish();
};
