
(clear-all)

(defvar *show-trace* t)

(defun show-trace () (setf *show-trace* t) (sgp :v t))
(defun hide-trace () (setf *show-trace* nil) (sgp :v nil))

(define-model SF-player

  ;;; tell it where the game server is
  (sgp-fct (list :sf-host *server-name*))
  #+sfapi (sgp-fct (list :sf-server-type *server-type*))
  (sgp :p-count t) 
  (sgp-fct (list :v *show-trace*))
  
  (sgp :er t :esc t :egs .0001 :ol t :lf .1  :ans .01 :rt -1 :pct nil :trace-detail low)
  ;; Temporal module's noise
  (eval `(sgp :sf-data-hook ,*data-hook*))
  (sgp :do-not-harvest game-state)
  
 
  (specify-compilation-buffer-type game-state perceptual) 

;; Temporal module's noise
  (sgp :TIME-NOISE .005)
  
 ;; Motor setup, timing, and randomization 
  (sgp :randomize-time 3)
  (sgp :dual-execution-stages t)
  (sgp :MOTOR-FEATURE-PREP-TIME 0.020)
  (sgp :MOTOR-INITIATION-TIME .020)
  (sgp :hold-until-timing .025)

  ;; visual configuration 
  (sgp :test-feats nil :visual-movement-tolerance 2 :do-not-harvest visual)

  (chunk-type gamestate x y orientation vx vy ship-alive status game  ;common
               offaim  moving-angle current-angle future-angle r1 r2 ship-speed   ;spacetrack
                time-to-outer thrust-angle vulnerability FORTRESS-ANGLE fortress-alive outward prop-to-big speed)   ;space fortess
  (chunk-type goal step state nextstep target aim game thrust-time;common
              rectangle  offaim  press-point stop-angle ship-speed ;spacetrack
              time-to-outer time-thresh dist-thresh vulnerability last-action outward prop-to-big speed) ;space fortess
  (chunk-type operator pre action arg1 result post  fail success)
  (chunk-type tracker rectangle death bad-aim good-aim
              hit miss badspeed goodspeed bighexdeath shelldeath smallhexdeath)
  (chunk-type mapping function key)

(default-chunks '(yes no speed adjust offaim turn-to-point test-dimension pressing-d pressing-key pressing-left down target-angle
                      THRUST-TO-TURNTAP-THRUST death rectangle press-point stop-angle tapthrust TAPPING-KEY pressing-a
                      CALCULATE-TURN-ANGLE maketurn2 test-speed  start done connected play start-playing-again do-step game-over
                      start-playing current-angle future-angle pressing-thrust tap-key thrust-to-turn ship-speed
                      NOFORTRESS FORTRESS-ALIVE TESTOBJECT DELAY? CHECK-DELAY ANGLE fortress-angle  double-space-bar
                             thrust thrust-angle increment-angle decrement-angle TIME-TO-OUTER space-bar shooting vulnerability? thrusting
                             double-spacebar state-changed reset hit time-thresh bighexdeath spacebar test-object prop-to-big
                             kill  shelldeath smallhexdeath thrust-time miss badspeed goodspeed))

(add-dm 


(w isa mapping function thrust key w)
(a isa mapping function counterclockwise key a)
(d isa mapping function clockwise key d)
)
(eval (cons 'add-dm *instructions*))


(sgp :ul t  :epl t :egs .05 :alpha .2 :iu 9  )  


(eval `(sgp :alpha ,*alpha*))
(eval `(sgp :initial-temp ,*factor*))

  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;; General productions for playing a SF-style game

    ;;; This production continues play at the very beginning


  (p get-started
     "Make sure game is running"
     ?game>
       game-state connected
     ?goal>
       buffer empty
     ?imaginal>
       state free
     ?game-state>
       state free
     ==>
     +goal>
         isa goal
         state start-playing
     +imaginal>
        isa tracker
     +game-state>
        isa gamestate)

    ;;; This production continues play for another game.
(p get-started-again
     "Make sure game is running"
     ?game>
       game-state connected
    =game-state>
       ship-alive yes 
       game =game
     =goal>
       state game-over
       game =game
==>
    =game-state>
      status nil
     =goal>
       state start-playing-again
       rectangle nil)

(p get-started-new-game
     "Make sure game is running"
     ?game>
       game-state connected
    =game-state>
       ship-alive yes 
       game =game
     =goal>
       state game-over
      - game =game
     ==>
     =goal>
         isa goal
         state start-playing
     +imaginal>
        isa tracker
     +game-state>
        isa gamestate)


;initiates for first game of space track
(p start-playing-spacetrack
     "once the ship is visible get going"
     =goal>
       isa goal
       state start-playing 
     -  game spacetrack 
    =game-state>
        game spacetrack
==>
     +tracker>
         control-slot aim
         good-slot rectangle
         bad-slot death
         min 3
         max 20.0
         bad-weight -1
     +tracker>
         control-slot ship-speed
         good-slot rectangle
         bad-slot death
         min 1.0
         max 4.0
         bad-weight -1
     +tracker>
         control-slot press-point
         good-slot rectangle
         bad-slot death
         min -25.0
         max 50.0
         bad-weight -1
     +tracker>
         control-slot stop-angle
         good-slot rectangle
         bad-slot death
         min 5.0
         max 40.0
         bad-weight -1
     +tracker>
         control-slot thrust-time
         good-slot rectangle
         bad-slot death
         min 0.067
         max 0.167
         bad-weight -1
    =goal> 
        state play
        offaim 25
        game spacetrack
     ;  thrust-time .09
     )

;initiates for first game of space fortress
 (p start-playing-spacefortress
     "once the ship is visible get going"
     =goal>
       isa goal
       state start-playing  
       - game spacefortress 
    =game-state>
        game spacefortress    
     ==>
     +tracker>
         control-slot prop-to-big
         good-slot kill
         bad-slot death
         min 0.1
         max 0.9
         bad-weight -1
     +tracker>
         control-slot time-thresh
         good-slot hit
         bad-slot reset
         min 9.0
         max 18.0
         bad-weight -1
     +tracker>
         control-slot time-to-outer 
         good-slot kill
         bad-slot death
         min 1.0
         max 5.0
         bad-weight -1
     +tracker>
         control-slot aim
         good-slot hit
         bad-slot miss
         min 3
         max 20.0
         bad-weight -1
     +tracker>
         control-slot thrust-angle
         good-slot goodspeed
         bad-slot badspeed
         min 85.0
         max 140.0
         bad-weight -1
     +tracker>
         control-slot thrust-time
         good-slot kill
         bad-slot death
         min 0.067
         max 0.167
         bad-weight -1
     =goal>
         state play
         speed 1.35
         vulnerability 9
         game spacefortress
         outward 0
    +temporal>
       isa time
     )


;applies at start of game or ship death  
(p start-playing-again-spacetrack
     =goal>
       isa goal
       state start-playing-again
       game spacetrack
     =game-state>
       ship-alive yes  
        game spacetrack
       r1 =rect
     ==>
   -retrieval>
     =goal>
         state play
        rectangle =rect
     )

  (p start-playing-again-spacefortress
     =goal>
       isa goal
       state start-playing-again
       game spacefortress
    =game-state>
        game spacefortress  
       ship-alive yes  
     ==>
   -retrieval>
     =goal>
         state play
     )

  (spp start-playing-spacetrack :u 100)
 (spp start-playing-spacefortress :u 100)
  (spp start-playing-again-spacetrack :u 100)
  (spp start-playing-again-spacefortress :u 100)

;;; This production determines game over and requires that information in game-state
 (p detect-trial-end
    "If a trial is clear things out"
    =goal>
      isa goal
    -  state game-over
    =game-state>
      status game-over
    ?manual>
      processor free
==>
    =goal>
      state game-over
      rectangle nil
      step nil
    +manual>
     isa release-all-fingers
    )
 (spp detect-trial-end :u 10000 ) 

  ;;; This production handles death
(p detect-destroyed
     =goal>
       isa goal
     - state start-playing-again
     =game-state>
       ship-alive no  
     ?manual>
       processor free
     ==>
 -retrieval>    
     =goal>
       state start-playing-again 
       step nil 
     +manual>
       isa release-all-fingers 
     )
  
  (spp detect-destroyed :u 90 )

 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;; General instruction interpretation productions
    (p do-something
       =goal>
          isa goal
          state play
      ?retrieval>
           state free
       ?MANUAL>
          state FREE
    =game-state>
        game =game 
      ==>
        =goal>
           state do-step
        +retrieval>
           isa operator
           pre =game
     )

    (p do-something-fast
       =goal>
          isa goal
          state play
          last-action shooting
      ?retrieval>
          ; buffer empty
           state free
       ?MANUAL>
          PROCESSOR FREE
    =game-state>
        game =game 
      ==>
        =goal>
           state do-step
           last-action nil
        +retrieval>
           isa operator
           pre =game
     )

(p object-present
   =goal>
      state do-step
    =retrieval>
      isa operator
      action test-object
      arg1 =object
      success =newstep
  =game-state>
     =object yes
     ==>
    +retrieval>
        isa operator
        pre =newstep)

(p object-not-present
   =goal>
      state do-step
    =retrieval>
      isa operator
      action test-object
     arg1 =object
      fail =newstep
  =game-state>
     =object no
     ==>
    +retrieval>
        isa operator
        pre =newstep)

(p test-dimension-success
   =goal>
      state do-step
      =dimension =thresh
    =retrieval>
      isa operator
      action test-dimension
      arg1 =dimension
      success =newstep
    =game-state>
        =dimension =val     
   !SAFE-EVAL! (> =val =thresh)
     ==>
    +retrieval>
        isa operator
        pre =newstep)

(p test-dimension-fail
   =goal>
      state do-step
      =dimension =thresh
    =retrieval>
      isa operator
      action test-dimension
      arg1 =dimension
      fail =newstep
    =game-state>
        =dimension =val     
   !SAFE-EVAL! (<= =val =thresh)
     ==>
    +retrieval>
        isa operator
        pre =newstep)




;productions for turning general to both games
(p turn-to-point1
   =goal>
      state do-step
    =retrieval>
      isa operator
      action turn-to-point
      arg1 =slot
      post =next
    =game-state>
      =slot =angle
     ==>
   =goal>
      nextstep =next
      state turn-to-point
       TARGET =angle)

(p skip-turn 
   =goal>
      state turn-to-point
      target =angle2
      aim =aim
      nextstep =next
      step nil
    =game-state>
      orientation =angle1
!safe-eval!   (< (angle-offset =angle2 =angle1) =aim)
     ==>
   +RETRIEVAL>
       PRE =next
=goal> 
      state do-step)

(p start-press-clockwise
   =goal>
      state turn-to-point
      target =angle2
      aim =aim
      step nil
    =game-state>
      orientation =angle1
  ?manual>
       state free 
!safe-eval!   (>= (mod (- =angle2 =angle1) 360) 180)
!safe-eval!   (>= (angle-offset =angle2 =angle1) 20)
     ==>
   +retrieval>
      isa mapping
      function clockwise
   =goal>
     step pressing-key)


(p press-clockwise-tap
   =goal>
      state turn-to-point
      target =angle2
      aim =aim
      step nil
    =game-state>
      orientation =angle1
  ?manual>
       state free 
!safe-eval!   (>= (mod (- =angle2 =angle1) 360) 180)
!safe-eval!   (>= (angle-offset =angle2 =angle1) =aim)
     ==>
   +retrieval>
      isa mapping
      function clockwise
   =goal>
     step tapping-key)



(p start-press-counterclockwise
   =goal>
      state turn-to-point
      target =angle2
      aim =aim
      step nil
    =game-state>
      orientation =angle1
  ?manual>
       state free 
!safe-eval! (< (mod (- =angle2 =angle1) 360) 180)
!safe-eval!   (>= (angle-offset =angle2 =angle1) 20)
     ==>
   +retrieval>
      isa mapping
      function counterclockwise
   =goal>
     step pressing-key)

(p press-counterclockwise-tap
   =goal>
      state turn-to-point
      target =angle2
      aim =aim
      step nil
    =game-state>
      orientation =angle1
  ?manual>
       state free 
!safe-eval! (< (mod (- =angle2 =angle1) 360) 180)
!safe-eval!   (>= (angle-offset =angle2 =angle1) =aim)
     ==>
   +retrieval>
      isa mapping
      function counterclockwise
   =goal>
     step tapping-key)



;productions that turn action intentions into specific finger actions that reflect game key binding
(p tap-key
   =goal>
      state do-step
   =RETRIEVAL>
       ACTION tap-key
       arg1 =function
     ==>
   +retrieval>
      isa mapping
      function =function
   =goal>
     state tap-key)


(p tap-w
   =goal>
      state tap-key
      thrust-time =time
   =RETRIEVAL>
       key w
    ?manual>
      state free
     ==>
   +MANUAL>
       CMD DELAYED-PUNCH
       FINGER MIDDLE
       HAND LEFT
       delay =time
   =goal>
      nextstep nil
      target nil
      state play)


(p tappping-d
   =goal>
      step tapping-key
   =RETRIEVAL>
       key d
    ?manual>
      state free
     ==>
   =goal>
     step nil
   +MANUAL>
       CMD DELAYED-PUNCH
       FINGER index
       HAND LEFT
       delay .09)

(p pressing-d
   =goal>
      step pressing-key
   =RETRIEVAL>
       key d
    ?manual>
      state free
     ==>
   =goal>
     step pressing-d
   +MANUAL>
       CMD hold-punch
       FINGER index
       HAND LEFT)

(p finish-press-d
   =goal>
      step pressing-d
      target =angle1
      nextstep =next
   =game-state>
     orientation =angle2
   ?manual-left>
      index down
  ?manual>
       preparation free 
!safe-eval! (<= (angle-offset =angle1 =angle2) 20)
     ==>
   +RETRIEVAL>
       PRE =next
   =goal>
      state do-step
      step nil
+manual>
  step nil
  cmd release
  hand left
  finger index )

(p pressing-a
   =goal>
      step pressing-key
   =RETRIEVAL>
       key a
    ?manual>
      state free
     ==>
   =goal>
     step pressing-a
   +MANUAL>
       CMD hold-punch
       FINGER ring
       HAND LEFT)

(p finish-press-a
   =goal>
      step pressing-a
      target =angle1
      nextstep =next
   =game-state>
     orientation =angle2
   ?manual-left>
      ring down
  ?manual>
       preparation free 
!safe-eval! (<= (angle-offset =angle1 =angle2) 20)
     ==>
   +RETRIEVAL>
       PRE =next
   =goal>
      state do-step
      step nil
+manual>
  step nil
  cmd release
  hand left
  finger ring )

(p tapping-a
   =goal>
      step tapping-key
   =RETRIEVAL>
       key a
    ?manual>
      state free
     ==>
   =goal>
     step nil
   +MANUAL>
       CMD DELAYED-PUNCH
       FINGER ring
       HAND LEFT
       delay .09)




;productions specific to space track
(p calculate-turn-angle
   =goal>
      state do-step
    =retrieval>
      isa operator
      action CALCULATE-TURN-ANGLE
      arg1 =slot
      post =next
    =game-state>
      =slot =angle
      moving-angle =moving-angle
    !SAFE-BIND! =TANGLE (TURN-ANGLE =ANGLE =moving-angle)
     ==>
   =goal>
      nextstep =next
      state turn-to-point
       TARGET =TANGLE)


(p thrust-new-rectangle
   =goal>
      state do-step
     press-point =distance
   =RETRIEVAL>
       ACTION thrust-to-turn
    =game-state>
      x =x
      y =y
      vx =vx
      vy =vy
      r2 =r2
    ?manual>
      state free
!safe-eval! (> (distance-in =r2 =x =y =vx =vy) =distance)
     ==>
   +MANUAL>
       CMD hold-PUNCH
       FINGER MIDDLE
       HAND LEFT
   =goal>
      rectangle =r2
      state pressing-thrust)

(p finish-thrust
   =goal>
      state pressing-thrust
      stop-angle =stop
   =game-state>
     current-angle =angle1
     moving-angle =angle2
   ?manual-left>
      middle down
  ?manual>
       preparation free 
!safe-eval! (<= (angle-offset =angle1 =angle2) =stop)
     ==>
   =goal>
      target nil
      nextstep nil
      state play
 +manual>
      cmd release
      hand left
      finger middle )


;productions specific to space fortress
(p increment-angle
   =goal>
      state do-step
      =dim =angle1
    =retrieval>
      isa operator
      action increment-angle
      arg1 =dim
      post =next
    =game-state>
        =dim =angle2
        orientation =orient    
 !safe-bind! =angle3 (mod (+ =orient (- =angle2 =angle1) 15) 360)
     ==>
   =GOAL>
       NEXTSTEP =NEXT
       STATE TURN-TO-POINT
       TARGET =ANGLE3)

(p decrement-angle
   =goal>
      state do-step
      =dim =angle1
    =retrieval>
      isa operator
      action decrement-angle
      arg1 =dim
      post =next
    =game-state>
        =dim =angle2
       orientation =orient  
!safe-bind! =angle3 (mod (+ =orient (- =angle2 =angle1) -15)  360)
     ==>
   =GOAL>
       NEXTSTEP =NEXT
       STATE TURN-TO-POINT
       TARGET =ANGLE3)


(p emergency-thrust
   =goal>
      isa goal
    ?manual>
      state free
  =game-state>
     < time-to-outer  .5
     ==>
   =goal>
      state play
   +MANUAL>
       CMD DELAYED-PUNCH
       FINGER MIDDLE
       HAND LEFT
       delay .18)

(spp emergency-thrust :u 10) 

(p delay-ok
   =goal>
      state do-step
      time-thresh =tick-thresh
   =RETRIEVAL>
       ACTION check-delay
       success =NEWSTEP
    =temporal>
      ticks =ticks
!safe-eval! (>= =ticks =tick-thresh)  
     ==>
  +RETRIEVAL>
       PRE =NEWSTEP)

(p press-spacebar
   =goal>
      state do-step
  =game-state>
      fortress-alive yes
    =retrieval>
      isa operator
      action spacebar
     ==>
    +temporal>
       isa time  
   =goal>
      state play
      last-action shooting
     +manual>
       isa delayed-punch
        hand right
       finger index
        delay .09)


(p double-spacebar
   =goal>
      state do-step
  =game-state>
      fortress-alive yes
    =retrieval>
      isa operator
      action double-spacebar
     ==>
   =goal>
      state double-spacebar
     +manual>
       isa delayed-punch
        hand right
       finger index
        delay .09)

(p double-shot
   =goal>
      state double-spacebar
   ?MANUAL>
       processor FREE
     ==>
     +manual>
       isa delayed-punch
      hand right
      finger index
       delay .09
    +temporal>
       isa time   
     =goal>
        state play)

(spp (finish-thrust press-spacebar double-spacebar) :reward 10)
(spp (GET-STARTED GET-STARTED-AGAIN START-PLAYING-AGAIN-spacetrack START-PLAYING-AGAIN-spacefortress
                  DETECT-DESTROYED DETECT-TRIAL-END  emergency-thrust) :reward t)


)


